from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Optional, cast

from rdkit import Chem
from rdkit.Chem import AllChem
from syntheseus.interface.bag import Bag
from syntheseus.interface.molecule import Molecule
from syntheseus.reaction_prediction.chem.utils import (
    molecule_bag_from_smiles,
    remove_atom_mapping_from_mol,
)

from retrochimera.chem.fixing import fix_mol
from retrochimera.chem.utils import (
    get_atom_mapping,
    map_removed_lhs_atoms_to_dummies,
    normalize_atom_mapping,
)


@dataclass
class RewriteResult:
    mols: Bag[Molecule]
    metadata: dict[str, Any] = field(default_factory=dict)


@dataclass(frozen=True, order=False)
class Rewrite:
    """Wrapper around a SMARTS string representing a graph rewrite."""

    lhs: str  # Left-hand side of the rewrite, represented as SMARTS.
    rhs: str  # Right-hand side of the rewrite, represented as SMARTS.
    mapping: list[tuple[int, int]]  # Atom mapping connecting both sides of the rewrite.

    # Other objects which encode the same information as above, just in a different form:
    smarts: str = field(repr=False, compare=False)
    rdkit_lhs_mol: Chem.Mol = field(repr=False, compare=False)
    rdkit_rhs_mol: Chem.Mol = field(repr=False, compare=False)
    rdkit_reaction: AllChem.ChemicalReaction = field(repr=False, compare=False)

    # Extra data to remember which rhs atoms are dummies (see below).
    _max_non_dummy_atom_mapping_num: int = field(repr=False, compare=False)

    @staticmethod
    def from_rdkit(template_smarts: str) -> Rewrite:
        template = AllChem.ReactionFromSmarts(template_smarts)
        normalize_atom_mapping(template)

        # If the rewrite removes some atoms present in the lhs, then this would prevent us from
        # being able to compute the localization below. We thus add dummies to make all the atoms
        # from the lhs appear in the rhs.
        template, max_atom_mapping_num = map_removed_lhs_atoms_to_dummies(template)

        # Make sure the template looks as expected.
        assert len(template.GetReactants()) == 1
        assert len(template.GetProducts()) == 1

        template_lhs_mol = Chem.Mol(template.GetReactantTemplate(0))
        template_rhs_mol = Chem.Mol(template.GetProductTemplate(0))

        mapping_num_to_id_lhs = get_atom_mapping(template_lhs_mol)
        mapping_num_to_id_rhs = get_atom_mapping(template_rhs_mol)

        template_mapping = [
            (mapping_num_to_id_lhs[id], mapping_num_to_id_rhs[id])
            for id in mapping_num_to_id_lhs.keys() & mapping_num_to_id_rhs.keys()
        ]

        for mol in [template_lhs_mol, template_rhs_mol]:
            remove_atom_mapping_from_mol(mol)

        return Rewrite(
            lhs=Chem.MolToSmarts(template_lhs_mol),
            rhs=Chem.MolToSmarts(template_rhs_mol),
            mapping=template_mapping,
            smarts=template_smarts,
            rdkit_lhs_mol=template_lhs_mol,
            rdkit_rhs_mol=template_rhs_mol,
            rdkit_reaction=template,
            _max_non_dummy_atom_mapping_num=max_atom_mapping_num,
        )

    def __getstate__(self) -> str:
        """Convert to easily serializable state, used by `pickle`."""
        return self.smarts

    def __setstate__(self, state: str) -> None:
        """Restore from state returned by `__getstate__`."""
        object.__setattr__(self, "__dict__", Rewrite.from_rdkit(state).__dict__)

    def __deepcopy__(self, memo) -> Rewrite:
        """Override `copy.deepcopy`s behaviour to never copy a `Rewrite` as it is immutable."""
        memo[id(self)] = self
        return self

    def apply(
        self, input: Molecule, fix_result: bool = True, return_localization: bool = True
    ) -> list[RewriteResult]:
        results: list[RewriteResult] = []
        result_to_index: dict[Bag[Molecule], int] = {}

        for raw_result in self.rdkit_reaction.RunReactants((input.rdkit_mol,)):
            # Make sure the result is a single molecule (should hold due to the asserts above).
            assert len(raw_result) == 1
            (result_rdkit_mol,) = raw_result

            localization: Optional[list[int]] = None
            atom_ids_to_remove: list[int] = []

            if return_localization:
                localization_tmp = [None] * self.rdkit_lhs_mol.GetNumAtoms()

                for atom in result_rdkit_mol.GetAtoms():
                    atom_props = atom.GetPropsAsDict()

                    if "old_mapno" in atom_props:
                        atom_mapping_idx = atom_props["old_mapno"]
                        atom_idx = self.mapping[atom_mapping_idx - 1][0]
                        localization_tmp[atom_idx] = atom_props["react_atom_idx"]

                        if atom_mapping_idx > self._max_non_dummy_atom_mapping_num:
                            atom_ids_to_remove.append(atom.GetIdx())

                assert None not in localization_tmp
                localization = cast(list[int], localization_tmp)

            if atom_ids_to_remove:
                # Remove dummy atoms that we may have added in `map_removed_lhs_atoms_to_dummies`;
                # proceed in reversed order to prevent IDs shifting from messing things up.
                result_rdkit_mol = Chem.RWMol(result_rdkit_mol)
                for atom_id in reversed(atom_ids_to_remove):
                    result_rdkit_mol.RemoveAtom(atom_id)

            if fix_result:
                smiles = fix_mol(result_rdkit_mol)
            else:
                try:
                    smiles = Chem.MolToSmiles(result_rdkit_mol)
                except Exception:
                    smiles = None

            if smiles is not None:
                mols = molecule_bag_from_smiles(smiles)
                if mols is not None:
                    index = result_to_index.get(mols)
                    if index is not None:
                        if return_localization:
                            current_localizations = results[index].metadata["localizations"]
                            assert localization is not None

                            current_localizations.append(localization)
                    else:
                        result_to_index[mols] = len(results)

                        if localization is not None:
                            metadata = {"localizations": [localization]}
                        else:
                            metadata = {}

                        results.append(RewriteResult(mols=mols, metadata=metadata))

        return results



from abc import ABCMeta, abstractmethod, abstractproperty
from dataclasses import dataclass
from importlib import import_module
from typing import Callable, Generic, Optional, TypeVar

from torch import Tensor, nn

from retrochimera.utils.misc import lookup_by_name


@dataclass
class EncoderOutput:
    """Encoder outputs for a batch of chemical objects (e.g. molecules).

    Attributes:
        atom_outputs: Encodings of individual atoms, as a concatenated tensor of shape
            `[total_num_atoms, atom_out_channels]`. Atoms from each molecule appear as consecutive
            segments, within each segment ordered as returned by `GetAtoms` from `rdkit`.
        bond_outputs: Encodings of individual bonds, as a concatenated tensor of shape
            `[2 * total_num_bonds, bond_out_channels]`. Bonds from each molecule appear as
            consecutive segments, within each segment ordered as returned by `GetBond` from `rdkit`.
            Each bond appears twice, in two directions (first the direction returned by `rdkit` and
            then the complementary one).
        mol_outputs: Encodings of entire molecules, as a tensor of shape
            `[batch_size, mol_out_channels]`.
    """

    atom_outputs: Optional[Tensor]
    bond_outputs: Optional[Tensor]
    mol_outputs: Optional[Tensor]


InputType = TypeVar("InputType")
DataType = TypeVar("DataType")
BatchType = TypeVar("BatchType")


class Encoder(nn.Module, Generic[InputType, DataType, BatchType], metaclass=ABCMeta):
    """Abstract interface for classes capable of encoding chemical objects.

    Generics:
        InputType: Type of the raw chemical entity to encode; should either be `Molecule` or another
            type that can also be understood in terms of atoms/bonds (also see `EncoderOutput`).
            Other examples include templates (graph rewrites) and reactions.
        DataType: Type of a single processed datapoint, for example a raw featurized molecular graph.
        BatchType: Type of a batch of processed datapoints, ready for processing by learned
            components of the encoder (if any).
    """

    @abstractproperty
    def preprocess(self) -> Callable[[InputType], DataType]:
        """Preprocess a raw input.

        Note: This function is a property returning a callable instead of a method to avoid the
        entirety of `self` being pickled when preprocessing is wrapped with multiprocessing. In this
        way, only the parts of `self` accessed in the implementation of this property will get
        attached to the function and sent over to the workers.
        """
        pass

    @abstractmethod
    def collate(self, inputs: list[DataType]) -> BatchType:
        pass

    @abstractmethod
    def forward(self, batch: BatchType) -> EncoderOutput:
        pass

    @abstractmethod
    def move_batch_to_device(self, batch: BatchType, device: str) -> BatchType:
        pass

    def forward_raw(self, inputs: list[InputType], device: str = "cpu") -> EncoderOutput:
        """Map raw inputs to final encoder outputs; useful for running inference."""
        batch = self.collate([self.preprocess(input) for input in inputs])
        return self.forward(self.move_batch_to_device(batch, device))

    @abstractproperty
    def atom_out_channels(self) -> Optional[int]:
        pass

    @abstractproperty
    def bond_out_channels(self) -> Optional[int]:
        pass

    @abstractproperty
    def mol_out_channels(self) -> Optional[int]:
        pass


def get_encoder_by_name(encoder_name: str) -> type[Encoder]:
    module = import_module("retrochimera.encoders." + encoder_name.removesuffix("Encoder").lower())
    return lookup_by_name(module, encoder_name)



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from syntheseus.cli import search

from retrochimera.cli.eval import BackwardModelConfig


@dataclass
class SearchConfig(BackwardModelConfig, search.BaseSearchConfig):
    """Config for running search for given search targets."""

    pass


def main(argv: Optional[list[str]]) -> None:
    search.main(argv, config_cls=SearchConfig)


if __name__ == "__main__":
    main(argv=None)



from torch import nn


class MLP(nn.Module):
    """Simple MLP with at least two layers and all hidden layers of the same size."""

    def __init__(
        self, in_dim, hidden_dim, out_dim, n_layers=1, activation=nn.ELU, dropout=0.2
    ) -> None:
        super(MLP, self).__init__()

        layers = [nn.Linear(in_dim, hidden_dim), nn.Dropout(dropout), activation()]

        for _ in range(n_layers):
            layers += [nn.Linear(hidden_dim, hidden_dim), nn.Dropout(dropout), activation()]

        layers += [nn.Linear(hidden_dim, out_dim)]

        self.mlp = nn.Sequential(*layers)

    def forward(self, x):
        return self.mlp(x)



import itertools
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import Callable, Iterable, Optional, Union

import numpy as np
from more_itertools import batched
from numpy.typing import NDArray
from syntheseus.interface.bag import Bag
from syntheseus.interface.molecule import Molecule
from syntheseus.reaction_prediction.utils.misc import parallelize

from retrochimera.chem.rewrite import RewriteResult
from retrochimera.chem.rule_application_server import RuleApplicationServer
from retrochimera.data.preprocessing.template_classification import get_unique_template_id
from retrochimera.data.template_reaction_sample import TemplateReactionSample
from retrochimera.encoders.base import Encoder
from retrochimera.encoders.featurizers import RawGraph


@dataclass
class ProcessedSample:
    input: RawGraph  # Input molecule featurized into a graph
    target: int  # ID of the ground-truth rewrite
    loc_target: NDArray  # Superposition of all correct matchings between rewrite lhs and input


def _process_sample(
    sample: tuple[Molecule, Bag[Molecule], int, list[RewriteResult]],
    input_preprocess_fn: Callable[[Molecule], RawGraph],
) -> Optional[tuple[RawGraph, int, list[int]]]:
    input, reactants, target, results = sample

    # Filter down to localizations that give the correct output.
    results = [result for result in results if result.mols == reactants]

    if not results:
        # TODO(krmaziar): This can happen due to chirality differences, possibly because of
        # issues with `rdchiral` and/or `RunReactions`. We skip these samples for now, but
        # should find a better solution later.
        return None

    assert len(results) == 1

    [result] = results
    localizations = result.metadata["localizations"]

    loc_target = np.zeros((len(localizations[0]), input.rdkit_mol.GetNumAtoms()), dtype=np.float32)
    for localization in localizations:
        for idx_rewrite, idx_input in enumerate(localization):
            loc_target[idx_rewrite][idx_input] += 1.0

    loc_target /= loc_target.sum(axis=1, keepdims=True)

    return input_preprocess_fn(input), target, loc_target


def preprocess_samples(
    samples: Iterable[TemplateReactionSample],
    rulebase_dir: Union[str, Path],
    input_encoder: Encoder,
    num_processes: int,
) -> Iterable[ProcessedSample]:
    # Extract the relevant info from each sample.
    samples_processed: Iterable = (
        (
            Molecule(smiles=sample.products_str),
            sample.reactants,
            get_unique_template_id(sample),
        )
        for sample in samples
    )

    server = RuleApplicationServer(
        rulebase_dir=rulebase_dir,
        num_processes=num_processes,
        rule_application_kwargs={"return_localization": True},
    )

    # Get rid of labels which are not in the `rulebase` (e.g. they were too rare).
    samples_processed = (sample for sample in samples_processed if sample[-1] in server.rule_ids)

    # Rerun templates to determine ground-truth localizations.
    samples_processed = itertools.chain.from_iterable(
        zip(
            samples_batch,
            server.apply_rules(
                inputs=[product for product, _, _, in samples_batch],
                rule_ids_to_apply=[[target] for _, _, target in samples_batch],
            ),
        )
        for samples_batch in batched(samples_processed, 4 * num_processes)
    )

    # Strip away superfluous nesting.
    samples_processed = (
        (input, reactants, target, results)
        for (input, reactants, target), [results] in samples_processed
    )

    samples_processed = parallelize(
        partial(_process_sample, input_preprocess_fn=input_encoder.preprocess),
        samples_processed,
        num_processes=num_processes,
    )

    for sample in samples_processed:
        # Skip samples where the rerunning did not produce the expected result.
        if sample is not None:
            yield ProcessedSample(*sample)



import torch
from torch import nn
from torch_geometric.nn.glob import GlobalAttention


class MultiheadAttentionalAggregation(nn.Module):
    """Multi-head global attention-based aggregation layer.

    Each head is implemented as a `torch_geometric.nn.glob.GlobalAttention` layer."""

    def __init__(self, in_channels: int, num_heads: int, out_channels: int) -> None:
        super().__init__()

        if out_channels % num_heads != 0:
            raise ValueError(
                f"Number of output channels {out_channels} has to be divisible by the number of attention heads {num_heads}"
            )

        out_channels_per_head = out_channels // num_heads
        self.heads = nn.ModuleList(
            [
                GlobalAttention(
                    gate_nn=nn.Linear(in_channels, 1, bias=False),
                    nn=nn.Linear(in_channels, out_channels_per_head),
                )
                for _ in range(num_heads)
            ]
        )
        self.out_channels = out_channels

    def forward(self, x: torch.Tensor, batch: torch.Tensor) -> torch.Tensor:
        return torch.cat([head(x, batch) for head in self.heads], dim=-1)



from __future__ import annotations

from typing import Any, Callable, Optional

import torch
import torch_geometric
from torch import nn
from torch_geometric.data import Batch, Data

import retrochimera.encoders
from retrochimera.encoders.base import Encoder, EncoderOutput, InputType
from retrochimera.encoders.featurizers import RawGraph, RawRewriteGraph
from retrochimera.layers.aggregation import MultiheadAttentionalAggregation
from retrochimera.layers.normalization import NormWrapper
from retrochimera.utils.misc import lookup_by_name


class GNNEncoder(Encoder[InputType, RawGraph, Batch]):
    """Implements an encoder backed by a GNN.

    This encoder depends on a featurizer that maps chemical objects to annotated graphs, such as
    those available in `dgllife.utils`. It can provide atom-, bond-, and molecule-level outputs:
    - atom-level outputs are directly produced by the GNN backbone
    - bond-level outputs are computed from atom-level outputs for endpoints and raw bond features
    - molecule-level outputs are computed by aggregating all atom-level outputs.
    """

    def __init__(
        self,
        atom_out_channels: int,
        bond_out_channels: Optional[int],
        mol_out_channels: int,
        gnn_class: str,
        gnn_kwargs: dict[str, Any],
        aggregation_num_heads: int,
        aggregation_dropout: float,
        featurizer_class: str,
        featurizer_kwargs: Optional[dict[str, Any]] = None,
        atom_categorical_features_channels: Optional[int] = None,
    ) -> None:
        super().__init__()

        # Handle the default value of `None`, which is here for backwards compatibility.
        featurizer_kwargs = featurizer_kwargs or {}

        self.featurizer = lookup_by_name(
            retrochimera.encoders.featurizers, featurizer_class
        ).get_default(**featurizer_kwargs)

        self._atom_out_channels = atom_out_channels
        self._bond_out_channels = bond_out_channels
        self._mol_out_channels = mol_out_channels

        num_total_atom_features = self.featurizer.num_atom_features
        if atom_categorical_features_channels is None:
            self.atom_categorical_features_emb = None
            num_total_atom_features += sum(self.featurizer.num_atom_categorical_features)
        else:
            self.atom_categorical_features_emb = nn.ModuleList(
                [
                    nn.Embedding(num_classes, atom_categorical_features_channels)
                    for num_classes in self.featurizer.num_atom_categorical_features
                ]
            )
            num_total_atom_features += atom_categorical_features_channels

        if gnn_class == "GPS_PNA":
            # Use the GPS layer from `graphium`.
            try:
                import graphium
            except ModuleNotFoundError:
                raise ModuleNotFoundError(
                    "You are using a model configuration that requires the `graphium` library. "
                    "Install it via `pip install syntheseus-graphium` or choose different settings."
                )

            if bond_out_channels is not None:
                raise ValueError("The GPS implementation does not support bond-level outputs")

            self.gnn = graphium.nn.architectures.FullGraphMultiTaskNetwork(
                gnn_kwargs={
                    "in_dim": gnn_kwargs["hidden_channels"],
                    "in_dim_edges": self.featurizer.num_bond_features,
                    "out_dim": atom_out_channels,
                    "hidden_dims": gnn_kwargs["hidden_channels"],
                    "depth": gnn_kwargs["num_layers"],
                    "activation": "gelu",
                    "last_activation": None,
                    "dropout": gnn_kwargs["dropout"],
                    "normalization": "layer_norm",
                    "last_normalization": "layer_norm",
                    "residual_type": gnn_kwargs.get("residual_type", "none"),
                    "virtual_node": gnn_kwargs.get("virtual_node"),
                    "layer_type": "pyg:gps",
                    "layer_kwargs": dict(
                        mpnn_type="pyg:pna-msgpass",
                        mpnn_kwargs={key: gnn_kwargs[key] for key in ["aggregators", "scalers"]},
                    ),
                },
                pre_nn_kwargs={
                    "in_dim": num_total_atom_features,
                    "out_dim": gnn_kwargs["hidden_channels"],
                    "hidden_dims": gnn_kwargs["hidden_channels"],
                    "depth": 1,
                    "activation": "relu",
                    "last_activation": "none",
                    "dropout": 0.0,
                },
            )

            self._norm_needs_batch_assignment = False
            self._use_graphium = True
        else:
            # Use a layer from `torch_geometric.nn.models`.
            norm = lookup_by_name(torch_geometric.nn.norm, gnn_kwargs["norm"])(
                in_channels=gnn_kwargs["hidden_channels"]
            )

            # We need to wrap some of the normalization layers, see `NormWrapper` for details.
            if gnn_kwargs["norm"] in NormWrapper.NORMS_REQUIRING_BATCH_ASSIGNMENT:
                self._norm_needs_batch_assignment = True
                norm = NormWrapper(norm)
            else:
                self._norm_needs_batch_assignment = False

            # Make a few adjustments to the kwargs.
            gnn_kwargs = {**gnn_kwargs, "norm": norm, "deg": torch.as_tensor(gnn_kwargs["deg"])}

            self.gnn = lookup_by_name(torch_geometric.nn.models, gnn_class)(
                in_channels=num_total_atom_features,
                out_channels=atom_out_channels,
                edge_dim=self.featurizer.num_bond_features,
                **gnn_kwargs,
            )
            self._use_graphium = False

        if bond_out_channels is None:
            self.edge_proj = None
        else:
            self.edge_proj = nn.Linear(
                2 * atom_out_channels + self.featurizer.num_bond_features,
                bond_out_channels,
            )

        self.aggregation = MultiheadAttentionalAggregation(
            in_channels=atom_out_channels,
            num_heads=aggregation_num_heads,
            out_channels=mol_out_channels,
        )
        self.aggregation_dropout = nn.Dropout(p=aggregation_dropout)

    @property
    def preprocess(self) -> Callable[[InputType], RawGraph]:
        return self.featurizer.__call__

    def collate(self, inputs: list[RawGraph]) -> Batch:
        return Batch.from_data_list([self.tensorize(input) for input in inputs])

    def tensorize(self, data: RawGraph) -> Data:
        if isinstance(data, RawRewriteGraph):
            extra_data = {"node_in_lhs": torch.as_tensor(data.node_in_lhs)}
        else:
            extra_data = {}

        if data.atom_categorical_features is not None:
            extra_data["categorical_feat"] = torch.as_tensor(
                data.atom_categorical_features, dtype=torch.long
            )

        return Data(
            feat=torch.as_tensor(data.atom_features, dtype=torch.float),
            edge_index=torch.as_tensor(data.bonds, dtype=torch.long).T,
            edge_feat=torch.as_tensor(data.bond_features, dtype=torch.float),
            **extra_data,
        )

    def forward(self, batch: Batch) -> EncoderOutput:
        if hasattr(batch, "categorical_feat"):
            combined_embeddings = None
            for idx, feat in enumerate(batch.categorical_feat.T):
                if self.atom_categorical_features_emb is None:
                    embeddings = nn.functional.one_hot(
                        feat, num_classes=self.featurizer.num_atom_categorical_features[idx]
                    )
                    embeddings = embeddings.float().to(batch.feat.device)
                else:
                    embeddings = self.atom_categorical_features_emb[idx](feat)

                if combined_embeddings is None:
                    combined_embeddings = embeddings
                else:
                    # Aggregate embeddings for all categorical features through a simple sum.
                    combined_embeddings += embeddings

            batch.feat = torch.cat([batch.feat, combined_embeddings], dim=-1)
        else:
            assert self.atom_categorical_features_emb is None

        if self._use_graphium:
            atom_outputs = self.gnn(batch).feat
        else:
            atom_features, edge_index, bond_features, batch_assignment = (
                batch.feat,
                batch.edge_index,
                batch.edge_feat,
                batch.batch,
            )

            if self._norm_needs_batch_assignment:
                for norm in self.gnn.norms:
                    norm.batch = batch_assignment

            atom_outputs = self.gnn(atom_features, edge_index, edge_attr=bond_features)

        if self.edge_proj is not None:
            # Compute edge outputs by concatenating endpoint outputs with original bond features...
            bond_outputs = torch.cat(
                (bond_features, atom_outputs[edge_index.T].reshape(bond_features.shape[0], -1)),
                dim=-1,
            )

            # ...and projecting into the desired dimensionality.
            bond_outputs = self.edge_proj(bond_outputs)
        else:
            bond_outputs = None

        if self.aggregation is not None:
            # Compute graph outputs by appropriately aggregating node outputs.
            mol_outputs = self.aggregation(atom_outputs, batch.batch)
            mol_outputs = self.aggregation_dropout(mol_outputs)
        else:
            mol_outputs = None

        return EncoderOutput(
            atom_outputs=atom_outputs, bond_outputs=bond_outputs, mol_outputs=mol_outputs
        )

    def move_batch_to_device(self, batch: Batch, device: str) -> Batch:
        return batch.to(device)

    @property
    def atom_out_channels(self) -> int:
        return self._atom_out_channels

    @property
    def bond_out_channels(self) -> Optional[int]:
        return self._bond_out_channels

    @property
    def mol_out_channels(self) -> int:
        return self._mol_out_channels



"""
MIT License

Copyright (c) 2017-Present OpenNMT

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Original Source: https://github.com/OpenNMT/OpenNMT-py/blob/v3.5.1/onmt/modules/multi_headed_attn.py
Modifications:
1. Removed the `self.parallel_gpu` attribute and corresponding logic from the `MultiHeadedAttention` class, including calls to `torch.distributed.all_reduce` and `torch.nn.utils.skip_init`.
"""
from importlib import import_module
from math import log, sqrt
from typing import Optional, Tuple

import torch
import torch.nn as nn
from torch import Tensor
from torch.nn.functional import scaled_dot_product_attention
from torch.utils.checkpoint import checkpoint

from retrochimera.opennmt.modules.alibi_position_bias import AlibiPositionalBias

# Help functions for Rotary Embeddings
# https://arxiv.org/pdf/2104.09864.pdf
# too convoluted to make maxseqlen a parameter.
# we suppose src_seq_len at training and max_length at inference
# are both < 2048 tokens.


def rotaryembeddings(dim: int, maxseqlen=2048, base=10000, device=None):
    inv_freq = 1.0 / (base ** (torch.arange(0, dim, 2).float() / dim))
    tmax = torch.arange(maxseqlen, device=inv_freq.device)
    rope = torch.outer(tmax, inv_freq).float()
    # rope is now matrix [maxseqlen, dim/2]
    rope = torch.polar(torch.ones_like(rope), rope)
    rope = torch.cat((rope, rope), dim=1)
    if device is not None:
        rope = rope.to(device)
    cos = rope[:, : rope.size(1) // 2].real.contiguous().half()
    sin = rope[:, : rope.size(1) // 2].imag.contiguous().half()
    return rope, cos, sin


def rotate_half(x):
    """Rotates half the hidden dims of the input."""
    x1 = x[..., : x.shape[-1] // 2]
    x2 = x[..., x.shape[-1] // 2 :]
    return torch.cat((-x2, x1), dim=-1)


def apply_rotary_emb(query, key, rope, interleave):
    if interleave:
        query = query.transpose(1, 2)
        key = key.transpose(1, 2)
        query_ = query.float().reshape(*query.shape[:-1], -1, 2)
        query_ = torch.view_as_complex(query_)
        key_ = key.float().reshape(*key.shape[:-1], -1, 2)
        key_ = torch.view_as_complex(key_)
        rope = rope[:, : rope.size(1) // 2].view(1, query_.size(1), 1, query_.size(3))
        query_out = torch.view_as_real(query_ * rope).flatten(3)
        key_out = torch.view_as_real(key_ * rope).flatten(3)
        return query_out.transpose(1, 2).type_as(query), key_out.transpose(1, 2).type_as(key)
    else:
        cos, sin = rope.real, rope.imag
        rotary_dim = cos.size(1)
        head_dim = query.size(3)
        if rotary_dim < head_dim:
            q_embed = (query[:, :, :, :rotary_dim] * cos) + (
                rotate_half(query[:, :, :, :rotary_dim]) * sin
            )
            k_embed = (key[:, :, :, :rotary_dim] * cos) + (
                rotate_half(key[:, :, :, :rotary_dim]) * sin
            )
            q_embed = torch.cat([q_embed, query[:, :, :, rotary_dim:]], dim=-1)
            k_embed = torch.cat([k_embed, key[:, :, :, rotary_dim:]], dim=-1)
        else:
            q_embed = (query * cos) + (rotate_half(query) * sin)
            k_embed = (key * cos) + (rotate_half(key) * sin)
        return q_embed.type_as(query), k_embed.type_as(key)


# Help functions for max_relative positions
# https://arxiv.org/abs/1803.02155


def relative_matmul(x: Tensor, z: Tensor, transpose: bool) -> Tensor:
    """
    Helper function for relative positions attention.
    https://arxiv.org/pdf/1803.02155.pdf
    x shape [batch_size x heads x q_len x k_len]
    """
    batch_size, heads, length, _ = x.size()
    x_t = x.permute(2, 0, 1, 3)
    x_t_r = x_t.contiguous().view(length, heads * batch_size, -1)
    if transpose:
        z = z.transpose(1, 2)
    x_tz_matmul = torch.matmul(x_t_r, z)
    x_tz_matmul_r = x_tz_matmul.view(length, batch_size, heads, -1)
    x_tz_matmul_r_t = x_tz_matmul_r.permute(1, 2, 0, 3)
    return x_tz_matmul_r_t


def gen_relative_positions(
    length: int,
    max_relative_positions: int,
    cache: bool = False,
    device: Optional[torch.device] = None,
) -> Tensor:
    """Generate the clipped relative positions matrix
    for a given length and maximum relative positions"""
    if cache:
        distance_mat = torch.arange(-length + 1, 1, 1, device=device).unsqueeze(0)
    else:
        range_vec = torch.arange(length, device=device)
        range_mat = range_vec.unsqueeze(-1).expand(-1, length).transpose(0, 1)
        distance_mat = range_mat - range_mat.transpose(0, 1)
    distance_mat_clipped = torch.clamp(
        distance_mat, min=-max_relative_positions, max=max_relative_positions
    )
    # Shift values to be >= 0
    final_mat = distance_mat_clipped + max_relative_positions
    return final_mat


def _relative_position_bucket(
    relative_position, bidirectional=True, num_buckets=32, max_distance=128
):
    """
    Adapted from Mesh Tensorflow:
    https://github.com/tensorflow/mesh/blob/0cb87fe07da627bf0b7e60475d59f95ed6b5be3d/
    mesh_tensorflow/transformer/transformer_layers.py#L593
    Translate relative position to a bucket number for relative attention.
    The relative position is defined as memory_position - query_position,
    i.e. the distance in tokens from the attending position to the attended-to
    position. If bidirectional=False, then positive relative positions are invalid.
    We use smaller buckets for small absolute relative_position and larger buckets for
    larger absolute relative_positions. All relative positions >=max_distance map to the
    same bucket. All relative positions <=-max_distance map to the same bucket.
    This should allow for more graceful generalization to longer sequences than the
    model has been trained on

    Args:
        relative_position: an int32 Tensor
        bidirectional: a boolean - whether the attention is bidirectional
        num_buckets: an integer
        max_distance: an integer

    Returns:
        a Tensor with the same shape as relative_position, containing int32 values
        in the range [0, num_buckets)
    """
    relative_buckets = 0
    if bidirectional:
        num_buckets //= 2
        relative_buckets += (relative_position > 0).to(torch.long) * num_buckets
        relative_position = torch.abs(relative_position)
    else:
        relative_position = -torch.min(relative_position, torch.zeros_like(relative_position))
    # now relative_position is in the range [0, inf)
    # half of the buckets are for exact increments in positions
    max_exact = num_buckets // 2
    is_small = relative_position < max_exact

    # The other half of the buckets are for logarithmically bigger bins in positions
    # up to max_distance
    relative_position_if_large = max_exact + (
        torch.log(relative_position.float() / max_exact)
        / log(max_distance / max_exact)
        * (num_buckets - max_exact)
    ).to(torch.long)
    relative_position_if_large = torch.min(
        relative_position_if_large,
        torch.full_like(relative_position_if_large, num_buckets - 1),
    )

    relative_buckets += torch.where(is_small, relative_position, relative_position_if_large)
    return relative_buckets


def compute_bias(
    query_length,
    key_length,
    is_decoder,
    max_relative_positions,
    relative_positions_buckets,
    device=None,
):
    """Compute binned relative position bias"""
    context_position = torch.arange(query_length, dtype=torch.long, device=device)[:, None]
    memory_position = torch.arange(key_length, dtype=torch.long, device=device)[None, :]
    relative_position = memory_position - context_position  # shape (query_length, key_length)
    relative_position_bucket = _relative_position_bucket(
        relative_position,  # shape (query_length, key_length)
        bidirectional=(not is_decoder),
        num_buckets=relative_positions_buckets,
        max_distance=max_relative_positions,
    )
    return relative_position_bucket


# Help functions to split model dim per head


def shape(x: Tensor, dim_per_head: int) -> Tensor:
    """
    Projection.
    [batchsize x length x modeldim]
    -> [batchsize x heads x length x dimperhead]
    """
    x_0, x_1, _ = x.size()
    return x.view(x_0, x_1, -1, dim_per_head).transpose(1, 2)


def unshape(x: Tensor) -> Tensor:
    """
    Compute context.
    [batchsize x heads x length x dimperhead]
    -> [batchsize x length x modeldim]
    """
    x_0, x_1, _, x_3 = x.size()
    return x.transpose(1, 2).contiguous().view(x_0, -1, x_1 * x_3)


class MultiHeadedAttention(torch.nn.Module):
    """Multi-Head Attention module from "Attention is All You Need"
    :cite:`DBLP:journals/corr/VaswaniSPUJGKP17`.

    Similar to standard `dot` attention but uses
    multiple attention distributions simulataneously
    to select relevant items.

    .. mermaid::

       graph BT
          A[key]
          B[value]
          C[query]
          O[output]
          subgraph Attn
            D[Attn 1]
            E[Attn 2]
            F[Attn N]
          end
          A --> D
          C --> D
          A --> E
          C --> E
          A --> F
          C --> F
          D --> O
          E --> O
          F --> O
          B --> O

    Also includes several additional tricks.

    Args:
       head_count (int): number of parallel heads
       model_dim (int): the dimension of keys/values/queries,
           must be divisible by head_count
       dropout (float): dropout parameter
       max_relative_positions (int): max relative positions
       attn_type: "self" or "context"
    """

    def __init__(
        self,
        head_count: int,
        model_dim: int,
        dropout: float = 0.1,
        is_decoder: bool = True,
        max_relative_positions: int = 0,
        relative_positions_buckets: int = 0,
        rotary_interleave: Optional[bool] = True,
        rotary_theta: float = 1e4,
        rotary_dim: int = 0,
        attn_type: Optional[str] = None,
        self_attn_type: Optional[str] = None,
        add_qkvbias=False,
        num_kv=0,
        use_ckpting=[],
    ) -> None:
        assert (
            model_dim % head_count == 0
        ), "Model dimension must be divisible by the number of heads"
        self.dim_per_head = model_dim // head_count
        super(MultiHeadedAttention, self).__init__()
        self.head_count = head_count
        self.num_kv = num_kv
        if num_kv == 0:
            self.linear_keys = nn.Linear(model_dim, model_dim, bias=add_qkvbias)
            self.linear_values = nn.Linear(model_dim, model_dim, bias=add_qkvbias)
        else:
            self.linear_keys = nn.Linear(
                model_dim, self.dim_per_head * self.num_kv, bias=add_qkvbias
            )
            self.linear_values = nn.Linear(
                model_dim, self.dim_per_head * self.num_kv, bias=add_qkvbias
            )

        self.linear_query = nn.Linear(model_dim, model_dim, bias=add_qkvbias)
        self.softmax = nn.Softmax(dim=-1)
        self.dropout = nn.Dropout(dropout)
        self.dropout_p = dropout
        self.final_linear = nn.Linear(model_dim, model_dim, bias=add_qkvbias)
        self.is_decoder = is_decoder
        self.max_relative_positions = max_relative_positions
        self.relative_positions_buckets = relative_positions_buckets
        self.attn_type = attn_type
        self.self_attn_type = self_attn_type
        self.layer_cache = (
            False,
            {"keys": torch.tensor([]), "values": torch.tensor([])},
        )
        if relative_positions_buckets > 0:
            self.relative_attention_bias = nn.Embedding(relative_positions_buckets, head_count)
            self.relative_positions_embeddings = None
        elif max_relative_positions > 0:
            # https://arxiv.org/pdf/1803.02155.pdf
            # in the paper they suggest either two embeds
            # relative_key / relative_value or only
            # relative_key. We implemented the same embed
            # for both.
            vocab_size = max_relative_positions * 2 + 1
            self.relative_positions_embeddings = nn.Embedding(vocab_size, self.dim_per_head)
            self.relative_attention_bias = None
        else:
            self.relative_positions_embeddings = None
            self.relative_attention_bias = None

            if max_relative_positions == -1:  # rotary embeddings
                if rotary_dim == 0:
                    self.rotary_dim = self.dim_per_head
                else:
                    self.rotary_dim = rotary_dim
                self.rope, self.cos, self.sin = rotaryembeddings(self.rotary_dim, base=rotary_theta)
                self.rotary_interleave = rotary_interleave
                self.rotary_theta = rotary_theta
            else:
                self.cos = None
                self.sin = None
                self.rotary_interleave = None
            if max_relative_positions == -2:  # alibi positional bias
                self.alibi = AlibiPositionalBias(head_count)

        self.maybe_ckpt = checkpoint if "mha" in use_ckpting else lambda f, x: f(x)

        try:
            flash_pack = import_module("flash_attn")
            if (
                hasattr(flash_pack, "flash_attn_func")
                and torch.cuda.get_device_capability()[0] >= 8
            ):
                self.flash_attn_func = getattr(flash_pack, "flash_attn_func")
                self.flash_attn_with_kvcache = getattr(flash_pack, "flash_attn_with_kvcache")
                self.flash2 = True
            else:
                self.flash2 = False
        except ImportError:
            self.flash2 = False

    def update_dropout(self, dropout: float) -> None:
        self.dropout.p = dropout
        self.dropout_p = dropout

    def forward(
        self,
        key: Tensor,
        value: Tensor,
        query: Tensor,
        mask: Optional[Tensor] = None,
        sliding_window: int = 0,
        step: int = 0,
        return_attn: Optional[bool] = False,
        self_attn_type: Optional[str] = None,
    ) -> Tuple[Tensor, Tensor]:
        """
        Compute the context vector and the attention vectors.

        Args:
           key (Tensor): set of `key_len`
               key vectors ``(batch, key_len, dim)``
           value (Tensor): set of `key_len`
               value vectors ``(batch, key_len, dim)``
           query (Tensor): set of `query_len`
               query vectors  ``(batch, query_len, dim)``
           mask: binary mask 1/0 indicating which keys have
               zero / non-zero attention ``(batch, query_len, key_len)``
           step (int): decoding step (used for Rotary embedding)
        Returns:
           (Tensor, Tensor):

           * output context vectors ``(batch, query_len, dim)``
           * Attention vector in heads ``(batch, head, query_len, key_len)``.
        """
        # 1) Project key, value, and query.
        # as a reminder at training layer_cache[0] remains False
        key_pad_mask = self.layer_cache[1].get("key_pad_mask", None)
        if self.layer_cache[0]:
            # Retrieve keys and values from the KV cache (decoding mode only).
            if self.attn_type == "self":
                query, key, value = (
                    self.linear_query(query),
                    self.linear_keys(query),
                    self.linear_values(query),
                )

                query = shape(query, self.dim_per_head)
                key = shape(key, self.dim_per_head)
                value = shape(value, self.dim_per_head)
                start_pos = step
                seqlen = query.size(2)

                if (
                    step == 0
                    or not self.flash2
                    or self.self_attn_type != "scaled-dot-flash"
                    or self.max_relative_positions not in [0, -1]
                    or query.size(0) > 128
                    or query.dtype != torch.float16
                ):
                    if self.max_relative_positions == -1:  # Rotary Embeddings
                        if seqlen + start_pos > self.rope.size(0):
                            # Resize rotary embeddings.
                            self.rope, self.cos, self.sin = rotaryembeddings(
                                self.rotary_dim,
                                maxseqlen=(seqlen + start_pos + 2048),
                                base=self.rotary_theta,
                                device=self.rope.device,
                            )
                        rope = self.rope[start_pos : start_pos + seqlen]
                        query, key = apply_rotary_emb(
                            query, key, rope, interleave=self.rotary_interleave
                        )

                    if self.layer_cache[1]["keys"].numel() != 0:
                        key = torch.cat((self.layer_cache[1]["keys"], key), dim=2)
                        value = torch.cat((self.layer_cache[1]["values"], value), dim=2)
                        if sliding_window > 0 and key.size(2) > sliding_window:
                            key = key[:, :, 1:, :]
                            value = value[:, :, 1:, :]

                    self.layer_cache[1]["keys"] = key
                    self.layer_cache[1]["values"] = value

                else:
                    if start_pos >= self.layer_cache[1]["keys"].size(2):
                        self.layer_cache[1]["keys"] = torch.cat(
                            [
                                self.layer_cache[1]["keys"],
                                torch.zeros(
                                    self.layer_cache[1]["keys"].shape[:-2]
                                    + (32,)
                                    + self.layer_cache[1]["keys"].shape[-1:],
                                    device=query.device,
                                ).half(),
                            ],
                            dim=-2,
                        )
                        self.layer_cache[1]["values"] = torch.cat(
                            [
                                self.layer_cache[1]["values"],
                                torch.zeros(
                                    self.layer_cache[1]["values"].shape[:-2]
                                    + (32,)
                                    + self.layer_cache[1]["values"].shape[-1:],
                                    device=query.device,
                                ).half(),
                            ],
                            dim=-2,
                        )
                        if self.max_relative_positions == -1 and start_pos + 32 >= self.rope.size(
                            0
                        ):
                            # Resize rotary embeddings.
                            # We take a margin of 32 tokens as the kv_cache
                            #   is incremented by 32 tokens every 32 tokens.
                            self.rope, self.cos, self.sin = rotaryembeddings(
                                self.rotary_dim,
                                maxseqlen=(start_pos + 2048),
                                base=self.rotary_theta,
                                device=self.rope.device,
                            )

                    if sliding_window > 0 and key.size(2) > sliding_window:
                        self.layer_cache[1]["keys"] = self.layer_cache[1]["keys"][:, :, 1:, :]
                        self.layer_cache[1]["values"] = self.layer_cache[1]["values"][:, :, 1:, :]
                    context = self.flash_attn_with_kvcache(
                        query.transpose(1, 2),
                        self.layer_cache[1]["keys"].transpose(1, 2),
                        self.layer_cache[1]["values"].transpose(1, 2),
                        key.transpose(1, 2),
                        value.transpose(1, 2),
                        rotary_cos=self.cos,
                        rotary_sin=self.sin,
                        cache_seqlens=step,
                        rotary_interleaved=self.rotary_interleave,
                    ).transpose(1, 2)
                    attn_output = self.final_linear(unshape(context))

                    return attn_output, None

            elif self.attn_type == "context":
                query = self.linear_query(query)
                query = shape(query, self.dim_per_head)
                if self.layer_cache[1]["keys"].numel() == 0:
                    key, value = self.linear_keys(key), self.linear_values(value)
                    key = shape(key, self.dim_per_head)
                    value = shape(value, self.dim_per_head)
                else:
                    key, value = (
                        self.layer_cache[1]["keys"],
                        self.layer_cache[1]["values"],
                    )
                self.layer_cache[1]["keys"] = key
                self.layer_cache[1]["values"] = value

            if key_pad_mask is not None:
                # Increase the cached key pad mask by concatenation.
                # For decoding only.
                if step > 0:
                    y = torch.zeros(
                        (key_pad_mask.size(0), key_pad_mask.size(1), 1),
                        dtype=torch.bool,
                        device=key_pad_mask.device,
                    )
                    self.layer_cache[1]["key_pad_mask"] = torch.cat((key_pad_mask, y), 2)
                    key_pad_mask = self.layer_cache[1]["key_pad_mask"]
        else:
            # Retrieve keys and values from linear layers (training mode).
            key = self.maybe_ckpt(self.linear_keys, key)
            value = self.maybe_ckpt(self.linear_values, value)
            query = self.maybe_ckpt(self.linear_query, query)

            key = shape(key, self.dim_per_head)
            value = shape(value, self.dim_per_head)
            query = shape(query, self.dim_per_head)

            if self.max_relative_positions == -1:  # Rotary Embeddings
                start_pos = 0
                seqlen = query.size(2)
                if seqlen > self.rope.size(0):
                    # Resize rotary embeddings.
                    self.rope, self.cos, self.sin = rotaryembeddings(
                        self.rotary_dim,
                        maxseqlen=(seqlen + 2048),
                        base=self.rotary_theta,
                        device=query.device,
                    )
                rope = self.rope[start_pos : start_pos + seqlen].to(query.device)
                query, key = apply_rotary_emb(query, key, rope, interleave=self.rotary_interleave)

        b, h, l, d = key.size()
        if self.num_kv > 0:
            qh = query.size(1)
            # expand key on heads dimension when it's less than query heads (multi-query variant)
            key = key.view(b, -1, 1, l, d).repeat(1, 1, qh // h, 1, 1)
            key = key.view(b, qh, l, d)

            # expand value on heads dimension when it's less than query heads (multi-query variant)
            value = value.view(b, -1, 1, l, d).repeat(1, 1, qh // h, 1, 1)
            value = value.view(b, qh, l, d)

        # 2) When standard pos. enc. or rotary, use flash attention

        # Ultimately flashv2 will be part of pytorch https://github.com/pytorch/pytorch/pull/105602
        # In the meantime: if vanilla tranformer or Rotary embeddings (not rel_pos, not alibi)
        # then use flash2 if seq len > 256 otherwise use xtransformer from pt2 uptream
        flash2 = self.flash2 and l > 256  # https://github.com/Dao-AILab/flash-attention/issues/591

        if (
            self.max_relative_positions in [-1, 0]
            and not return_attn
            and query.device != torch.device("cpu")
            and self.self_attn_type == "scaled-dot-flash"
        ):
            # Apply flash2 attention.
            causal = self.is_decoder and self.attn_type == "self" and mask is not None
            if self.is_decoder and self.attn_type == "self" and flash2:
                if causal:
                    window_size = (-1, -1) if sliding_window == 0 else (sliding_window, 0)
                else:
                    window_size = (-1, -1)
                attn_output = self.flash_attn_func(
                    query.transpose(1, 2),
                    key.transpose(1, 2),
                    value.transpose(1, 2),
                    dropout_p=self.dropout_p,
                    causal=causal,
                    window_size=window_size,
                ).transpose(1, 2)
            else:
                # Apply scaled dot product attention.
                with torch.backends.cuda.sdp_kernel(
                    enable_flash=False, enable_math=True, enable_mem_efficient=True
                ):
                    attn_output = scaled_dot_product_attention(
                        query,
                        key,
                        value,
                        ~mask if mask is not None else None,
                        self.dropout_p,
                        is_causal=causal,
                    )
            attn = None

        else:
            query /= sqrt(self.dim_per_head)
            # batch x num_heads x query_len x key_len
            scores = torch.matmul(query, key.transpose(2, 3))

            if self.relative_attention_bias is not None:
                q_len = key.size(2) if self.layer_cache[0] else query.size(2)
                relative_position_bucket = compute_bias(
                    q_len,
                    key.size(2),
                    self.is_decoder,
                    self.max_relative_positions,
                    self.relative_positions_buckets,
                    device=key.device,
                )
                values = self.relative_attention_bias(
                    relative_position_bucket
                )  # shape (query_length, key_length, num_heads)
                position_bias = values.permute([2, 0, 1]).unsqueeze(
                    0
                )  # shape (1, num_heads, query_length, key_length)
                if self.layer_cache[0]:
                    position_bias = position_bias[:, :, -query.size(2) :, :]
                scores.add_(position_bias)

            elif self.relative_positions_embeddings is not None:
                key_len = key.size(2)
                # 1 or key_len x key_len
                relative_positions_matrix = gen_relative_positions(
                    key_len,
                    self.max_relative_positions,
                    cache=self.layer_cache[0],
                    device=key.device,
                )
                #  1 or key_len x key_len x dim_per_head
                relations_keys = self.relative_positions_embeddings(relative_positions_matrix)
                scores.add_(relative_matmul(query, relations_keys, True))
            elif self.max_relative_positions == -2:  # Alibi
                scores = self.alibi(scores)

            scores = scores.float()
            if key_pad_mask is not None and mask is None:
                mask = key_pad_mask.unsqueeze(1)

            if mask is not None:
                # not 100% necessary but expand to nb of heads
                mask = mask.expand(-1, self.head_count, -1, -1)
                # now mask and scores have the same shape
                scores = scores.masked_fill(mask, -1e18)

            # 3) Apply attention dropout and compute context vectors.
            attn = self.softmax(scores).to(query.dtype)
            drop_attn = self.dropout(attn) if self.dropout_p > 0 else attn

            attn_output = torch.matmul(drop_attn, value)

            if self.relative_positions_embeddings is not None:
                # We use the same embeddings for key and value
                relations_values = relations_keys
                attn_output.add_(relative_matmul(drop_attn, relations_values, False))

        context = unshape(attn_output)
        if key_pad_mask is not None:
            if key_pad_mask.size(0) > 1 and context.size(1) > 1:
                x = key_pad_mask.squeeze(1).unsqueeze(2).expand(-1, -1, context.size(2))
                context = context.masked_fill(x, 0)

        if self.layer_cache[0]:
            attn_output = self.final_linear(context)
        else:
            attn_output = self.maybe_ckpt(self.final_linear, context)

        return attn_output, attn



import math
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, Optional, Union

import torch
from more_itertools import batched
from torch import nn
from torch.nn.init import xavier_uniform_, zeros_
from torch.optim.lr_scheduler import OneCycleLR

from retrochimera.chem.rules import RuleBase
from retrochimera.data.preprocessing.smiles_transformer import ProcessedSample, preprocess_samples
from retrochimera.data.smiles_reaction_sample import SmilesReactionSample
from retrochimera.data.smiles_tokenizer import Tokenizer
from retrochimera.models.lightning import AbstractModel
from retrochimera.opennmt.modules.transformer_decoder import TransformerDecoder
from retrochimera.opennmt.modules.transformer_encoder import TransformerEncoder
from retrochimera.utils.logging import get_logger
from retrochimera.utils.pytorch import FuncLR

logger = get_logger(__name__)


@dataclass
class Batch:
    encoder_input: torch.Tensor  # (batch_size, src_len), dtype=torch.int
    encoder_padding_mask: torch.Tensor  # (batch_size, src_len), dtype=torch.bool, True/False
    decoder_input: torch.Tensor  # (batch_size, tgt_len-1), dtype=torch.int
    decoder_padding_mask: torch.Tensor  # (batch_size, tgt_len-1), dtype=torch.bool, True/False
    target: torch.Tensor  # (batch_size, tgt_len-1), dtype=torch.int
    target_mask: torch.Tensor  # (batch_size, tgt_len-1), dtype=torch.bool, True/False


class SmilesTransformerModel(AbstractModel[SmilesReactionSample, ProcessedSample, Batch]):
    """Transformer model operating on product/reactants in SMILES form."""

    DEFAULT_VOCAB_FILE_NAME = "vocab.txt"

    def __init__(
        self,
        vocab_path: str,
        hidden_dim: int = 256,
        n_layers: int = 6,
        n_heads: int = 8,
        feedforward_dim: int = 2048,
        activation: str = "gelu",
        max_seq_len: int = 512,
        dropout: float = 0.1,
        positional_encoding_type: str = "SinusoidalInterleaved",
        label_smoothing: float = 0.0,
        share_encoder_decoder_input_embedding: bool = True,
        initialization: str = "xavier",
        add_qkvbias: bool = False,
        layer_norm: str = "standard",
        num_kv: int = 0,
        parallel_residual: bool = False,
        schedule: str = "noam",
        warm_up_steps: int = 8000,
        n_classes: int = 0,
        total_steps: int = 100000,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)

        # Build the tokenizer.
        logger.info(f"Loaded tokenizer from {vocab_path}")
        self.tokenizer = Tokenizer.from_vocab_file(vocab_path)
        self.vocab_size = len(self.tokenizer)
        self.pad_token_id = self.tokenizer.pad_token_id
        self.begin_token_id = self.tokenizer.begin_token_id
        self.end_token_id = self.tokenizer.end_token_id

        # Model hyperparameters.
        self.hidden_dim = hidden_dim
        self.n_layers = n_layers
        self.n_heads = n_heads
        self.feedforward_dim = feedforward_dim
        self.activation = activation
        self.dropout = dropout
        self.positional_encoding_type = positional_encoding_type
        self.label_smoothing = label_smoothing
        self.share_encoder_decoder_input_embedding = share_encoder_decoder_input_embedding
        self.initialization = initialization
        self.add_qkvbias = add_qkvbias
        self.layer_norm = layer_norm
        self.num_kv = num_kv
        self.parallel_residual = parallel_residual

        self.max_seq_len = max_seq_len
        self.n_classes = n_classes  # number of classes for template prediction
        self.total_steps = total_steps  # for some lr schedulers
        self.dropout_layer = nn.Dropout(dropout)

        # 1. Token embedding and positional encoding.
        if self.share_encoder_decoder_input_embedding:
            self.token_embedding = nn.Embedding(
                self.vocab_size, self.hidden_dim, padding_idx=self.pad_token_id
            )
        else:
            self.encoder_token_embedding = nn.Embedding(
                self.vocab_size, self.hidden_dim, padding_idx=self.pad_token_id
            )
            self.decoder_token_embedding = nn.Embedding(
                self.vocab_size, self.hidden_dim, padding_idx=self.pad_token_id
            )

        self.register_buffer(
            "positional_encoding", self.absolute_positional_encoding()
        )  # fixed sinusoidal APE (Vaswani et al., 2017)

        # 2. Encoder and decoder.
        self.encoder = TransformerEncoder(
            num_layers=self.n_layers,
            d_model=self.hidden_dim,
            heads=self.n_heads,
            d_ff=self.feedforward_dim,
            dropout=self.dropout,
            attention_dropout=self.dropout,
            pos_ffn_activation_fn=self.activation,
            add_qkvbias=self.add_qkvbias,
            add_ffnbias=True,
            layer_norm=self.layer_norm,  # "standard" or "rms"
            num_kv=self.num_kv,
            parallel_residual=self.parallel_residual,
        )

        self.decoder = TransformerDecoder(
            num_layers=self.n_layers,
            d_model=self.hidden_dim,
            heads=self.n_heads,
            d_ff=self.feedforward_dim,
            dropout=self.dropout,
            attention_dropout=self.dropout,
            pos_ffn_activation_fn=self.activation,
            add_qkvbias=self.add_qkvbias,
            add_ffnbias=True,
            layer_norm=self.layer_norm,  # "standard" or "rms"
            num_kv=self.num_kv,
            parallel_residual=self.parallel_residual,
        )

        self.token_fc = nn.Linear(self.hidden_dim, self.vocab_size)
        self.log_softmax = nn.LogSoftmax(dim=2)
        self.loss_fn = nn.CrossEntropyLoss(
            reduction="none", ignore_index=self.pad_token_id, label_smoothing=self.label_smoothing
        )

        self._init_params()

        # Training hyperparameters.
        self.learning_rate = kwargs["learning_rate"]
        self.optimizer_betas = kwargs["optimizer_betas"]
        self.schedule = schedule
        self.warm_up_steps = warm_up_steps  # for some lr schedulers

        # Inference/sampling hyperparameters.
        self.num_beams = 10
        self.bad_token_log_prob = -1e5

    @property
    def hyperparameters_excluded_from_checkpoint(self) -> list[str]:
        return ["vocab_path"]

    def forward(self, batch_input: Batch) -> dict[str, torch.Tensor]:
        """Run a batch of inputs through the model.

        Args:
            batch_input: Batch from `collate` which contains the following attributes:
                encoder_input: Integer tensor of token IDs with shape `(batch_size, src_len)`.
                encoder_padding_mask: Boolean tensor of padding masks with shape
                    `(batch_size, src_len)`.
                decoder_input: Integer tensor of decoder token IDs with shape
                    `(batch_size, tgt_len-1)`.
                decoder_padding_mask: Boolean tensor of decoder padding masks with shape
                    `(batch_size, tgt_len-1)`.
                target: Unused in this function.
                target_mask: Unused in this function.

        Returns:
            Dictionary with keys "token_output" and "model_output".
        """
        # Encoder
        encode_dict = {
            "encoder_input": batch_input.encoder_input,  # shape: (batch_size, src_len)
            "encoder_padding_mask": batch_input.encoder_padding_mask,  # shape: (batch_size, src_len)
        }
        memory = self.encode(encode_dict)  # shape: (batch_size, src_len, hidden_dim)

        self.decoder.init_state(
            batch_input.encoder_input.transpose(0, 1).contiguous().unsqueeze(2), None, None
        )

        # Decoder
        decode_dict = {
            "decoder_input": batch_input.decoder_input,
            "decoder_padding_mask": batch_input.decoder_padding_mask,
            "memory_input": memory,
            "memory_padding_mask": batch_input.encoder_padding_mask.clone(),
        }
        model_output, token_output, token_log_probs = self.decode(decode_dict)
        output = {
            "model_output": model_output,
            "token_output": token_output,
            "token_log_probs": token_log_probs,
        }

        return output

    def ttv_step(self, batch: Batch, step_name: str):
        """Training, validation, and test step."""
        if step_name == "train":
            self.train()
        else:
            self.eval()

        batch_size = batch.encoder_input.shape[0]

        model_output = self.forward(batch)
        loss = self.calculate_loss(batch, model_output)
        with torch.no_grad():
            token_acc = self.calculate_token_accuracy(batch, model_output)
            perplexity = self.calculate_perplexity(batch, model_output)

        self.log(
            f"{step_name}_loss",
            loss,
            prog_bar=True,
            on_step=True,
            sync_dist=True,
            batch_size=batch_size,
        )
        self.log(
            f"{step_name}_token_acc",
            token_acc,
            prog_bar=True,
            on_step=True,
            sync_dist=True,
            batch_size=batch_size,
        )
        self.log(
            f"{step_name}_perplexity",
            perplexity,
            prog_bar=True,
            on_step=True,
            sync_dist=True,
            batch_size=batch_size,
        )

        return loss

    def construct_input(
        self, token_ids: torch.Tensor, step: Optional[int] = None, is_encoder: bool = True
    ) -> torch.Tensor:
        """Construct an input embedding for a given token id tensor.

        Args:
            token_ids: Integer tensor of token IDs with shape (batch_size, seq_len) or (batch_size, seq_len, 1).
            step: Step for positional encoding.
            is_encoder: Whether the input is for the encoder or decoder.

        Returns:
            embs: Tensor of input embeddings with shape `(batch_size, seq_len, hidden_dim)`.
        """
        if token_ids.dim() == 3:  # the case of beam search for decoder input
            assert token_ids.size(2) == 1
            token_ids = token_ids.squeeze(2)
        elif token_ids.dim() != 2:
            raise ValueError(
                f"Expected token_ids to have shape (batch_size, seq_len) or (batch_size, seq_len, 1), got {token_ids.shape}"
            )

        _, seq_len = tuple(token_ids.size())

        if (
            self.share_encoder_decoder_input_embedding
        ):  # shared token embedding for encoder and decoder
            token_embs = self.token_embedding(token_ids)
        else:  # separate embedding for encoder and decoder
            if is_encoder:
                token_embs = self.encoder_token_embedding(
                    token_ids
                )  # shape: (batch_size, seq_len, hidden_dim)
            else:
                token_embs = self.decoder_token_embedding(
                    token_ids
                )  # shape: (batch_size, seq_len, hidden_dim)

        assert token_embs.size(1) == seq_len and token_embs.size(2) == self.hidden_dim

        token_embs = token_embs * math.sqrt(
            self.hidden_dim
        )  # scaling the embeddings like this is done in other transformer libraries

        # Start from step.
        step = step or 0
        assert (
            self.positional_encoding.shape[0] >= step + token_embs.shape[1]
        ), f"Sequence is {step + token_embs.shape[1]} but PositionalEncoding is {self.positional_encoding.shape[0]}. See max_len argument."

        positional_embs = self.positional_encoding[step : step + seq_len, :].unsqueeze(
            0
        )  # shape: (1, seq_len, hidden_dim)
        embs = token_embs + positional_embs  # shape: (batch_size, seq_len, hidden_dim)
        embs = self.dropout_layer(embs)

        return embs

    def encode(self, batch: dict[str, torch.Tensor]) -> torch.Tensor:
        """Construct a memory embedding for an encoder input.

        Args:
            batch: Dictionary containing two keys:
                "encoder_input": Integer tensor of token IDs with shape (batch_size, src_len)
                "encoder_padding_mask": Boolean tensor of padding masks with shape (batch_size, src_len)

        Returns:
            Tensor of memory embedding with shape `(src_len, batch_size, hidden_dim)`.
        """
        encoder_input = batch["encoder_input"]  # shape: (batch_size, src_len)
        encoder_padding_mask = batch["encoder_padding_mask"]  # shape: (batch_size, src_len)
        encoder_embs = self.construct_input(
            encoder_input, is_encoder=True
        )  # shape: (batch_size, src_len, hidden_dim)
        model_output = self.encoder.forward(
            src=encoder_embs, src_key_padding_mask=encoder_padding_mask
        )  # shape: (batch_size, src_len, hidden_dim)

        return model_output

    def decode(
        self,
        batch: dict[str, torch.Tensor],
        start_pos: int = 0,
        beam_idx: int = 0,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Construct an output from a given decoder input.

        Args:
            batch: Dictionary containing four keys:
                "decoder_input": Integer tensor of decoder token IDs with shape (batch_size, tgt_len-1)
                "decoder_padding_mask": Boolean tensor of decoder padding masks with shape (batch_size, tgt_len-1)
                "memory_input": Integer memory embedding tensor with shape (batch_size, src_len, hidden_dim)
                "memory_padding_mask": Boolean tensor of memory padding masks with shape (batch_size, src_len)
            start_pos: Starting position of decoder input for attention caching.

        Returns:
            Three decoder output tensors, each with shape (tgt_len-1, batch_size, hidden_dim),
            corresponding to raw output, token logits, and token log-probabilities, respectively.
        """
        decoder_input = batch["decoder_input"]
        decoder_padding_mask = batch["decoder_padding_mask"]  # shape: (batch_size, tgt_len-1)
        memory_input = batch["memory_input"]  # shape: (batch_size, src_len, hidden_dim)
        memory_padding_mask = batch["memory_padding_mask"]  # shape: (batch_size, src_len)

        decoder_embs = self.construct_input(
            decoder_input, is_encoder=False
        )  # shape: (batch_size, tgt_len-1, hidden_dim)

        _, seq_len, _ = tuple(decoder_embs.size())

        model_output, _ = self.decoder.forward(
            tgt=decoder_embs,
            tgt_key_padding_mask=decoder_padding_mask,
            enc_out=memory_input,
            enc_key_padding_mask=memory_padding_mask,
        )  # shape: (batch_size, tgt_len-1, hidden_dim)
        token_output = self.token_fc(model_output)  # shape: (batch_size, tgt_len - 1, hidden_dim)
        assert (
            token_output.size(1) == seq_len and token_output.size(2) == self.vocab_size
        ), f"{token_output.size()}"

        token_log_probs = self.log_softmax(
            token_output
        )  # shape: (batch_size, tgt_len - 1, vocab_size)

        return model_output, token_output, token_log_probs

    def absolute_positional_encoding(self) -> torch.Tensor:
        """Construct a tensor of positional embeddings from sine/cosine waves of varying wavelength.

        Returns:
            Tensor of shape `(self.max_seq_len, self.hidden_dim)` filled with positional embeddings.
        """
        if self.hidden_dim % 2 != 0:
            raise ValueError(
                f"Cannot use sin/cos positional encoding with odd dim {self.hidden_dim}"
            )

        if self.positional_encoding_type == "SinusoidalInterleaved":
            pe = torch.zeros(self.max_seq_len, self.hidden_dim)  # shape: (max_seq_len, hidden_dim)
            position = torch.arange(0, self.max_seq_len).unsqueeze(1)  # shape: (max_seq_len, 1)
            div_term = torch.exp(
                (
                    torch.arange(0, self.hidden_dim, 2, dtype=torch.float)
                    * -(math.log(10000.0) / self.hidden_dim)
                )
            )  # shape: (hidden_dim // 2)
            pe[:, 0::2] = torch.sin(
                position.float() * div_term
            )  # shape: (max_seq_len, hidden_dim//2)
            pe[:, 1::2] = torch.cos(
                position.float() * div_term
            )  # shape: (max_seq_len, hidden_dim//2)
        elif self.positional_encoding_type == "SinusoidalConcat":
            half_dim = self.hidden_dim // 2
            pe = math.log(10000) / (half_dim - 1)
            pe = torch.exp(
                torch.arange(half_dim, dtype=torch.float) * -pe
            )  # shape: (hidden_dim//2)
            pe = torch.arange(self.max_seq_len, dtype=torch.float).unsqueeze(1) * pe.unsqueeze(
                0
            )  # shape: (max_seq_len, hidden_dim//2)
            pe = torch.cat([torch.sin(pe), torch.cos(pe)], dim=1).view(
                self.max_seq_len, -1
            )  # shape: (max_seq_len, hidden_dim)
        else:
            raise ValueError(
                "Position encoding should be SinusoidalInterleaved or SinusoidalConcat."
            )
        return pe

    def compute_probs(
        self, reaction_smiles: list[str], minibatch_size: int = 32
    ) -> tuple[list[float], list[float]]:
        """Compute probabilities for a list of reaction SMILES strings.

        Args:
            reaction_smiles: List of reaction SMILES strings, where each reaction SMILES string has the format:
                reactant1.reactant2>reagent1.reagent2>product1.product2
        Returns:
            Lists of probabilities (total and average) of the reactions.
        """
        return_total_probs = []
        return_avg_probs = []
        for reaction_smiles_minibatch in batched(reaction_smiles, minibatch_size):
            # 1. Obtain the product and reactant SMILES strings in the minibatch.
            product_smi = [smi.split(">")[2] for smi in reaction_smiles_minibatch]
            reactants_smi = [smi.split(">")[0] for smi in reaction_smiles_minibatch]

            # 2. Tokenize the product and reactant SMILES strings and their masks.
            products_token_ids, products_masks = self.tokenizer.encode(
                product_smi, pad=True, right_padding=True
            )
            reactants_token_ids, reactants_masks = self.tokenizer.encode(
                reactants_smi, pad=True, right_padding=True, add_begin_token=True
            )

            products_token_ids = torch.tensor(products_token_ids)
            products_masks = torch.tensor(products_masks, dtype=torch.bool)
            reactants_token_ids = torch.tensor(reactants_token_ids)
            reactants_masks = torch.tensor(reactants_masks, dtype=torch.bool)

            batch = Batch(
                encoder_input=products_token_ids.to(self.device),
                encoder_padding_mask=products_masks.to(self.device),
                decoder_input=reactants_token_ids[:, :-1].to(self.device),
                decoder_padding_mask=reactants_masks[:, :-1].to(self.device),
                target=reactants_token_ids.clone()[:, 1:].to(self.device),
                target_mask=reactants_masks.clone()[:, 1:].to(self.device),
            )

            model_output = self.forward(batch)
            total_probs, avg_probs = self.calculate_probs(batch, model_output)
            return_total_probs.extend(total_probs.tolist())
            return_avg_probs.extend(avg_probs.tolist())

        return return_total_probs, return_avg_probs

    def calculate_loss(
        self, batch_input: Batch, model_output: dict[str, torch.Tensor]
    ) -> torch.Tensor:
        """Calculate the loss function to update the model.

        Args:
            batch_input: Input given to the model.
            model_output: Output from the model.

        Returns:
            Resulting loss.
        """

        tokens = batch_input.target.transpose(0, 1)
        padding_mask = batch_input.target_mask.transpose(0, 1)
        token_output = model_output["token_output"].transpose(0, 1)

        return self._calculate_mask_loss(token_output, tokens, padding_mask)

    def _calculate_mask_loss(
        self, token_output: torch.Tensor, target: torch.Tensor, target_mask: torch.Tensor
    ) -> torch.Tensor:
        """Calculate the loss function.

        Args:
            token_output: Tensor of shape `(tgt_len-1, batch_size, vocab_size)` containing token
                outputs from the transformer.
            target: Tensor of shape `(tgt_len-1, batch_size)` containing original (unmasked) SMILES
                token IDs from the tokenizer.
            target_mask: Tensor of shape `(tgt_len-1, batch_size)` containing padding mask for
                target tokens.

        Returns:
            Loss computed using cross-entropy.
        """

        seq_len, batch_size = tuple(target.size())
        token_pred = token_output.reshape((seq_len * batch_size, -1)).float()
        loss = self.loss_fn(token_pred, target.reshape(-1)).reshape((seq_len, batch_size))

        inv_target_mask = ~(target_mask > 0)
        num_tokens = inv_target_mask.sum()
        loss = loss.sum() / num_tokens

        return loss

    def calculate_probs(
        self, batch_input: Batch, model_output: dict[str, torch.Tensor]
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Calculate the probabilities (including total probability and average probability of a target sequence)."""
        target_ids = batch_input.target.transpose(0, 1)
        target_mask = batch_input.target_mask.transpose(0, 1)
        token_log_probs = model_output["token_log_probs"].transpose(0, 1)

        inv_target_mask = ~(target_mask > 0)
        log_probs = token_log_probs.gather(2, target_ids.unsqueeze(2)).squeeze(2)
        log_probs = log_probs * inv_target_mask

        total_log_probs = log_probs.sum(dim=0)
        total_probs = torch.exp(total_log_probs)

        num_tokens = inv_target_mask.sum(dim=0)
        probs = torch.exp(log_probs) * inv_target_mask
        probs = probs.sum(dim=0)
        avg_probs = probs / num_tokens.float()

        return total_probs, avg_probs

    def calculate_perplexity(
        self, batch_input: Batch, model_output: dict[str, torch.Tensor]
    ) -> torch.Tensor:
        """Calculate the perplexity."""
        target_ids = batch_input.target.transpose(0, 1)  # shape: (tgt_len-1, batch_size)
        target_mask = batch_input.target_mask.transpose(0, 1)  # shape: (tgt_len-1, batch_size)
        token_log_probs = model_output["token_log_probs"].transpose(
            0, 1
        )  # shape: (tgt_len-1, batch_size, vocab_size)

        inv_target_mask = ~(target_mask > 0)
        log_probs = token_log_probs.gather(2, target_ids.unsqueeze(2)).squeeze(2)
        log_probs = log_probs * inv_target_mask
        log_probs = log_probs.sum(dim=0)  # shape: (batch_size,)

        seq_lengths = inv_target_mask.sum(dim=0)  # shape: (batch_size,)
        perp = torch.exp(-log_probs / seq_lengths)

        return perp.mean()

    def calculate_token_accuracy(
        self, batch_input: Batch, model_output: dict[str, torch.Tensor]
    ) -> torch.Tensor:
        """Calculate the token accuracy."""
        token_ids = batch_input.target.transpose(0, 1)  # shape: (tgt_len-1, batch_size)
        target_mask = batch_input.target_mask.transpose(0, 1)  # shape: (tgt_len-1, batch_size)
        token_output = model_output["token_output"].transpose(
            0, 1
        )  # shape: (tgt_len-1, batch_size, vocab_size)

        target_mask = ~(target_mask > 0)
        _, pred_ids = torch.max(token_output.float(), dim=2)
        correct_ids = torch.eq(token_ids, pred_ids)
        correct_ids = correct_ids * target_mask

        num_correct = correct_ids.sum().float()
        total = target_mask.sum().float()
        accuracy = num_correct / total

        return accuracy

    def configure_optimizers(self):
        """Configure the optimizer and learning rate scheduler."""
        logger.info("Configuring optimizer")
        logger.info(f"Learning rate: {self.learning_rate}")
        logger.info(f"Betas: {self.optimizer_betas}")
        logger.info(f"Schedule: {self.schedule}")

        optimizer = torch.optim.Adam(
            self.parameters(), lr=self.learning_rate, betas=self.optimizer_betas
        )

        if self.schedule == "noam":
            logger.info("Using original transformer schedule (noam)")
            lr_scheduler = FuncLR(optimizer, lr_lambda=self._noam_lr)
        elif self.schedule == "cosine":
            logger.info("Using cosine LR schedule")
            lr_scheduler = FuncLR(optimizer, lr_lambda=self._warmup_cosine_lr)
        elif self.schedule == "constant":
            logger.info("Using constant LR schedule")
            lr_scheduler = FuncLR(optimizer, lr_lambda=self._constant_lr)
        elif self.schedule == "cycle":
            logger.info("Using cyclical LR schedule")
            lr_scheduler = OneCycleLR(optimizer, self.learning_rate, total_steps=self.total_steps)
        else:
            raise ValueError(f"Unknown schedule {self.schedule}")
        sch = {"scheduler": lr_scheduler, "interval": "step"}

        return [optimizer], [sch]

    def _constant_lr(self, step: int) -> float:
        if step < self.warm_up_steps:
            return (self.learning_rate / int(self.warm_up_steps)) * step

        return self.learning_rate

    def _noam_lr(self, step: int) -> float:
        mult = self.hidden_dim**-0.5
        step = 1 if step == 0 else step  # Stop div by zero errors
        lr = min(step**-0.5, step * (self.warm_up_steps**-1.5))

        return self.learning_rate * mult * lr

    def _warmup_cosine_lr(self, step: int, linear_decay: bool = False) -> float:
        if step < self.warm_up_steps:
            return (self.learning_rate / int(self.warm_up_steps)) * step
        else:
            if linear_decay:
                return max(
                    0.0,
                    float(self.total_steps - step)
                    / float(max(1.0, self.total_steps - self.warmup_num_steps)),
                )

            return 0.5 * (
                1.0
                + math.cos(
                    math.pi
                    * float(step - self.warmup_num_steps)
                    / float(max(1.0, self.total_steps - self.warmup_num_steps))
                )
            )

    def _init_params(self) -> None:
        """Apply Xavier uniform initialisation of learnable weights."""

        # Analyze model parameters.
        logger.info(f"Total parameter size: {sum(p.numel() for p in self.parameters())}")
        for name, param in self.named_parameters():
            logger.debug(f"{name}: {param.size()}")

        # Initialize model parameters.
        if self.initialization == "xavier":
            logger.info("Using Xavier uniform initialization")
            for p in self.parameters():
                if p.dim() > 1:
                    xavier_uniform_(p)
        elif self.initialization == "opennmt_xavier":
            logger.info("Using OpenNMT-py Xavier uniform initialization")
            for param_name, param in self.named_parameters():
                if param_name.endswith("weight") and param.dim() > 1:
                    xavier_uniform_(param)
                elif param_name.endswith("bias"):
                    zeros_(param)
        else:
            raise ValueError(f"Unknown initialization method: {self.initialization}")

    def collate(self, samples: list[ProcessedSample]) -> Batch:
        """Collate a list of samples into a batch."""
        products_token_ids, products_masks = self.tokenizer.pad_token_ids_list(
            [sample.encoded_products for sample in samples], right_padding=True
        )
        products_token_ids = torch.tensor(products_token_ids)
        products_masks = torch.tensor(products_masks, dtype=torch.bool)

        decoder_token_ids, decoder_masks = self.tokenizer.pad_token_ids_list(
            [sample.encoded_reactants for sample in samples], right_padding=True
        )
        decoder_token_ids = torch.tensor(decoder_token_ids)
        decoder_masks = torch.tensor(decoder_masks, dtype=torch.bool)

        return Batch(
            encoder_input=products_token_ids,  # shape: (batch_size, src_len)
            encoder_padding_mask=products_masks,  # shape: (batch_size, src_len)
            decoder_input=decoder_token_ids[:, :-1],  # shape: (batch_size, tgt_len-1)
            decoder_padding_mask=decoder_masks[:, :-1],  # shape: (batch_size, tgt_len-1)
            target=decoder_token_ids.clone()[:, 1:],  # shape: (batch_size, tgt_len-1)
            target_mask=decoder_masks.clone()[:, 1:],  # shape: (batch_size, tgt_len-1)
        )

    def preprocess(
        self, samples: Iterable[SmilesReactionSample], num_processes: int = 0
    ) -> Iterable[ProcessedSample]:
        yield from preprocess_samples(
            samples=samples,
            rulebase_dir=self.rulebase_dir,
            tokenizer=self.tokenizer,
            num_processes=num_processes,
        )

    def set_rulebase(self, rulebase: RuleBase, rulebase_dir: Union[str, Path]) -> None:
        # Inherited from other template-based models (Not used in this model now)
        self.rulebase_dir = rulebase_dir



import os

import pytest

from retrochimera.chem.rules import RuleBase
from retrochimera.data.dataset import DataFold, InMemoryReactionDataset, ReactionDataset
from retrochimera.data.template_reaction_sample import (
    TemplateApplicationResult,
    TemplateReactionSample,
)


def pytest_configure(config):
    # Set env var to disable wandb during tests (using `wandb.init` would not work for subprocesses)
    os.environ["WANDB_MODE"] = "disabled"


@pytest.fixture
def tiny_rulebase() -> RuleBase:
    rulebase = RuleBase()
    for smarts in [
        "[C:1][C:2][C:3][C:4]>>([C:1][C:2].[C:3][C:4])",
        "[C:1][C:2][C:3][C:4]>>([C:1][C:2][C:3].[C:4])",
    ]:
        rulebase.add_rule(smarts=smarts)

    return rulebase


@pytest.fixture
def rulebase() -> RuleBase:
    rulebase = RuleBase()
    for smarts in [
        "([#6A&h0X3v4+0:20]-[#7A&h1X3v3+0:22])>>([#6A&h0X3v4+0:20]-[#8A&h1X2v2+0:38].[#7A&h2X3v3+0:22])",
        "([#8A&h1X2v2+0:1])>>([#8A&h0X2v2+0:1]-[#6A&h3X4v4+0:16])",
        "([#7A&h2X3v3+0:21])>>([#7A&h0X3v4+:21](=[#8A&h0X&v2+0:25])-[#8A&h0X&v-:26])",
        "([#8A&h0X2v2+0:10]-[#6A&h2X4v4+0:11])>>([#6A&h2X4v4+0:11]-[BrA&h0X&v+0:19].[#8A&h1X2v2+0:10])",
    ]:
        # Value for `n_support` is based on `raw_reaction_data` below.
        rulebase.add_rule(smarts=smarts, n_support=7)

    return rulebase


@pytest.fixture
def rulebase_uneven() -> RuleBase:
    rulebase = RuleBase()
    for idx, smarts in enumerate(
        [
            "([#6A&h0X3v4+0:20]-[#7A&h1X3v3+0:22])>>([#6A&h0X3v4+0:20]-[#8A&h1X2v2+0:38].[#7A&h2X3v3+0:22])",
            "([#8A&h1X2v2+0:1])>>([#8A&h0X2v2+0:1]-[#6A&h3X4v4+0:16])",
            "([#7A&h2X3v3+0:21])>>([#7A&h0X3v4+:21](=[#8A&h0X&v2+0:25])-[#8A&h0X&v-:26])",
            "([#8A&h0X2v2+0:10]-[#6A&h2X4v4+0:11])>>([#6A&h2X4v4+0:11]-[BrA&h0X&v+0:19].[#8A&h1X2v2+0:10])",
        ]
    ):
        # Value for `n_support` is based on `raw_reaction_data_uneven` below.
        rulebase.add_rule(smarts=smarts, n_support=7 - idx)

    return rulebase


@pytest.fixture
def raw_reaction_data() -> list[tuple[str, int]]:
    return [
        ("CCN.O=C(O)c1cccs1>>CCNC(=O)c1cccs1", 0),
        ("CCCCCCN.NCCC(=O)O>>CCCCCCNC(=O)CCN", 0),
        ("CN.O=C(O)c1ccccc1I>>CNC(=O)c1ccccc1I", 0),
        ("Nc1ccccc1.O=C(O)CCl>>O=C(CCl)Nc1ccccc1", 0),
        ("CN.O=C(O)c1cc[nH]c1>>CNC(=O)c1cc[nH]c1", 0),
        ("N#CCC(=O)O.Nc1ccccc1>>N#CCC(=O)Nc1ccccc1", 0),
        ("CC(C)(C)N.O=C(O)CCBr>>CC(C)(C)NC(=O)CCBr", 0),
        ("COCCCCC(=O)OC>>COCCCCC(=O)O", 1),
        ("C=C(C)C(=O)OC>>C=C(C)C(=O)O", 1),
        ("COc1cccc(Cl)c1N>>Nc1c(O)cccc1Cl", 1),
        ("COc1ccc(S)cc1Cl>>Oc1ccc(S)cc1Cl", 1),
        ("COc1cnc(Cl)nc1C>>Cc1nc(Cl)ncc1O", 1),
        ("COc1ccc2ccsc2c1>>Oc1ccc2ccsc2c1", 1),
        ("COC(=O)Cc1ccccc1>>O=C(O)Cc1ccccc1", 1),
        ("O=[N+]([O-])c1ccn[nH]1>>Nc1ccn[nH]1", 2),
        ("COc1ncccc1[N+](=O)[O-]>>COc1ncccc1N", 2),
        ("CNc1ccccc1[N+](=O)[O-]>>CNc1ccccc1N", 2),
        ("O=Cc1ccccc1[N+](=O)[O-]>>Nc1ccccc1C=O", 2),
        ("CCCn1ccc([N+](=O)[O-])n1>>CCCn1ccc(N)n1", 2),
        ("Nc1ccc(F)cc1[N+](=O)[O-]>>Nc1ccc(F)cc1N", 2),
        ("O=[N+]([O-])c1ccn(CCO)n1>>Nc1ccn(CCO)n1", 2),
        ("CCCCBr.OCCCO>>CCCCOCCCO", 3),
        ("C#CCBr.COCCOCCO>>C#CCOCCOCCOC", 3),
        ("C#CCBr.OCC1CCCC1>>C#CCOCC1CCCC1", 3),
        ("CCCCBr.OCc1ccsc1>>CCCCOCc1ccsc1", 3),
        ("CCCCCCO.O=C(O)CBr>>CCCCCCOCC(=O)O", 3),
        ("O=C(O)CBr.OC1CCC1>>O=C(O)COC1CCC1", 3),
        ("BrCCCCBr.OC1CCCC1>>BrCCCCOC1CCCC1", 3),
    ]


@pytest.fixture
def raw_reaction_data_uneven(raw_reaction_data: list[tuple[str, int]]) -> list[tuple[str, int]]:
    rule_to_num_samples_left = [7, 6, 5, 4]

    data = []
    for reaction_smiles, target in raw_reaction_data:
        if rule_to_num_samples_left[target] > 0:
            rule_to_num_samples_left[target] -= 1
            data.append((reaction_smiles, target))

    return data


def build_reaction_datset(raw_data: list[tuple[str, int]], rulebase: RuleBase) -> ReactionDataset:
    data = []
    for reaction_smiles, target in raw_data:
        data.append(
            TemplateReactionSample.from_reaction_smiles_strict(
                reaction_smiles=reaction_smiles,
                mapped=False,
                template_application_results=[TemplateApplicationResult(template_id=target)],
            )
        )

    return InMemoryReactionDataset(samples={fold: data for fold in DataFold}, rulebase=rulebase)


@pytest.fixture
def reaction_dataset(
    raw_reaction_data: list[tuple[str, int]], rulebase: RuleBase
) -> ReactionDataset:
    return build_reaction_datset(raw_reaction_data, rulebase)


@pytest.fixture
def reaction_dataset_uneven(
    raw_reaction_data_uneven: list[tuple[str, int]], rulebase_uneven: RuleBase
) -> ReactionDataset:
    return build_reaction_datset(raw_reaction_data_uneven, rulebase_uneven)



import math
from typing import Any, Optional, Sequence

from syntheseus import Molecule, SingleProductReaction
from syntheseus.interface.reaction import ReactionMetaData
from syntheseus.reaction_prediction.inference_base import ExternalBackwardReactionModel
from syntheseus.reaction_prediction.utils.inference import get_unique_file_in_dir

from retrochimera.chem.rewrite import RewriteResult
from retrochimera.chem.rules import RuleBasedRetrosynthesizer, RulePrediction


class TemplateLocalizationModel(RuleBasedRetrosynthesizer, ExternalBackwardReactionModel):
    """Wrapper for the template localization model."""

    def __init__(
        self,
        *args,
        localization_score_weight: float = 2.25,
        classification_temperature: Optional[float] = 30.0,
        output_temperature: float = 1.0,
        **kwargs,
    ) -> None:
        """Initializes the TemplateLocalization model wrapper.

        Assumed format of the model directory:
        - `model_dir` contains the model checkpoint as the only `*.ckpt` file
        """
        import torch

        from retrochimera.models.template_localization import (
            TemplateLocalizationModel as LocalizationModel,
        )

        super().__init__(*args, **kwargs)
        self.start_server(rulebase_dir=self.model_dir)

        self.model = LocalizationModel.load_from_checkpoint(
            get_unique_file_in_dir(self.model_dir, pattern="*.ckpt")
        )
        self._localization_score_weight = localization_score_weight
        self._classification_temperature = classification_temperature
        self._output_temperature = output_temperature

        self.model.to(self.device)
        self.model.eval()

        # These should be set if the model was loaded from a trained checkpoint.
        assert self.model.all_rewrites_atom_outputs is not None
        assert self.model.all_rewrites_batch_idx is not None

        # Unpack the rewrite atom outputs for convenience.
        self.all_rewrites_atom_outputs_list: list[torch.Tensor] = []
        for rule_id in range(self.model.n_classes):
            self.all_rewrites_atom_outputs_list.append(
                self.model.all_rewrites_atom_outputs[self.model.all_rewrites_batch_idx == rule_id]
            )

    @property
    def _rule_application_server_kwargs(self) -> dict[str, Any]:
        return {"return_localization": True}

    def get_model_info(self) -> dict[str, Any]:
        return dict(self.model.hparams)

    def get_parameters(self):
        return self.model.parameters()

    def _predict_ranked_rules(
        self, targets: list[Molecule], top_k: int = 50
    ) -> list[list[RulePrediction]]:
        import torch

        from retrochimera.utils.pytorch import get_sorted_ids_and_probs, tensor_to_list

        # Preprocess and batch the inputs.
        input_graphs = self.model.input_encoder.collate(
            [self.model.input_encoder.preprocess(target) for target in targets]
        )
        input_graphs = self.model.input_encoder.move_batch_to_device(input_graphs, self.device)
        input_graphs_enc = self.model.input_encoder.forward(input_graphs)

        assert input_graphs_enc.mol_outputs is not None
        assert self.model.all_rewrites_mol_outputs is not None

        # Run the classification branch to get logits for all templates.
        batch_rule_logits = self.model.forward_classification(
            input_reprs=input_graphs_enc.mol_outputs,
            rewrite_reprs=self.model.all_rewrites_mol_outputs,
            rewrite_ids=torch.arange(self.model.n_classes, device=self.device),
            temperature=self._classification_temperature,
        )

        input_graphs_atom_outputs = input_graphs_enc.atom_outputs
        assert input_graphs_atom_outputs is not None

        results = []
        for idx, (rule_ids, rule_probs) in enumerate(
            get_sorted_ids_and_probs(batch_rule_logits, k=top_k)
        ):
            input_graph_atom_outputs = torch.t(input_graphs_atom_outputs[input_graphs.batch == idx])
            graph_results = []

            for rule_id, rule_prob in zip(rule_ids, rule_probs):
                localization_scores = torch.mm(
                    self.all_rewrites_atom_outputs_list[rule_id], input_graph_atom_outputs
                )
                localization_scores = torch.nn.functional.log_softmax(localization_scores, dim=-1)

                graph_results.append(
                    RulePrediction(
                        id=rule_id,
                        prob=rule_prob,
                        localization_scores=tensor_to_list(localization_scores),
                    )
                )

            results.append(graph_results)

        return results

    def _fill_prediction_probability(self, results_metadata: list[ReactionMetaData]) -> None:
        import torch

        scaled_scores = [
            metadata["combined_score"] / self._output_temperature for metadata in results_metadata  # type: ignore[typeddict-item]
        ]
        probabilities = torch.nn.functional.softmax(torch.as_tensor(scaled_scores), dim=-1)

        for metadata, probability in zip(results_metadata, probabilities.numpy().tolist()):
            metadata["probability"] = probability

    def _rerank_results(self, results: list[RewriteResult]) -> None:
        for rewrite_result in results:
            classification_score = math.log(rewrite_result.metadata["template_probability"])
            localization_score = score_localization(rewrite_result)
            combined_score = (
                classification_score + self._localization_score_weight * localization_score
            )

            rewrite_result.metadata.update(
                {
                    "classification_score": classification_score,
                    "localization_score": localization_score,
                    "combined_score": combined_score,
                }
            )

        results.sort(key=lambda t: t.metadata["combined_score"], reverse=True)

    def _get_reactions(
        self, inputs: list[Molecule], num_results: int
    ) -> list[Sequence[SingleProductReaction]]:
        return [
            prediction_list[:num_results]
            for prediction_list in self.predict(inputs, top_k=num_results)
        ]


def score_localization(result: RewriteResult) -> float:
    localizations = result.metadata["localizations"]
    localization_scores = result.metadata["localization_scores"]

    total_logprob = 0.0
    for index_rewrite_lhs, indices_input_mol in enumerate(zip(*localizations)):
        prob = 0.0
        for index_input_mol in set(indices_input_mol):
            prob += math.exp(localization_scores[index_rewrite_lhs][index_input_mol])

        total_logprob += math.log(prob)

    # A sum of log probabilities is more mathematically correct, but an average empirically seems
    # to work a bit better, so we do the latter.
    return total_logprob / len(localizations[0])



import random
import weakref
from collections import deque
from multiprocessing import Pipe, Process
from multiprocessing.connection import Connection
from pathlib import Path
from typing import Any, Deque, Optional, Union

from more_itertools import divide
from syntheseus.interface.molecule import Molecule
from syntheseus.reaction_prediction.utils.misc import cpu_count, suppress_rdkit_outputs

from retrochimera.chem.rewrite import RewriteResult
from retrochimera.chem.rules import RuleBase, get_products
from retrochimera.utils.logging import get_logger

logger = get_logger(__name__)


def _worker_process(
    rulebase_dir: Union[str, Path], rule_ids: list[int], connection: Connection
) -> None:
    rulebase = RuleBase.load_from_file(dir=rulebase_dir, rule_ids=rule_ids)
    rule_application_kwargs = connection.recv()

    # Send a dummy signal when the rulebase has been loaded.
    connection.send(None)

    with suppress_rdkit_outputs():
        while True:
            request = connection.recv()

            if request is None:
                break
            else:
                smiles, rule_ids = request

                # SMILES comes from a `Molecule` object in the main process, so it is already canonical.
                mol = Molecule(smiles, canonicalize=False)

                for rule_id in rule_ids:
                    connection.send(
                        get_products(mol, rule=rulebase[rule_id].rxn, **rule_application_kwargs)
                    )

    connection.close()


class RuleApplicationServer:
    """Server that supports highly parallel application of reaction rules."""

    def __init__(
        self,
        rulebase_dir: Union[str, Path],
        num_processes: int = cpu_count(),
        rule_application_kwargs: Optional[dict[str, Any]] = None,
    ) -> None:
        """Initialize the server.

        Args:
            rulebase_dir: Directory containing the rulebase file which will be accessed by the
                parallel processes. The constructor will block until all the processes have loaded
                up the rulebase and are ready to process requests. The rulebase file may be read
                again later if one of the processes hangs or fails and has to be restarted.
            num_processes: Number of parallel processes to spawn. The rules will be (roughly)
                equally divided between them.
            rule_application_kwargs: Extra arguments to pass through to `get_products`.
        """
        # First, *randomly* split the rules into `num_processes` groups to assign to each process.
        self.rule_ids = RuleBase.load_rule_ids_from_file(dir=rulebase_dir)

        rule_ids_shuffled = list(self.rule_ids)
        random.shuffle(rule_ids_shuffled)

        self._rule_id_to_process_id: dict[int, int] = {}
        self._rule_ids_per_process: list[list[int]] = [[] for _ in range(num_processes)]

        for process_id, rule_ids in enumerate(divide(num_processes, rule_ids_shuffled)):
            for rule_id in rule_ids:
                self._rule_id_to_process_id[rule_id] = process_id
                self._rule_ids_per_process[process_id].append(rule_id)

        self._rulebase_dir = rulebase_dir
        self._num_processes = num_processes
        self._rule_application_kwargs = rule_application_kwargs or {}

        self._running = True
        self._connections: dict[int, Connection] = {}
        self._processes: dict[int, Process] = {}

        self._start_processes(list(range(num_processes)))

        for process_id in range(num_processes):
            # Block until the process has loaded its chunk of the rulebase.
            self._connections[process_id].recv()

        # Register the `close` method to be called when the server object is garbage collected.
        weakref.finalize(self, self.close)

    def _start_processes(self, process_ids: list[int]) -> None:
        for process_id in process_ids:
            assert process_id not in self._processes
            assert process_id not in self._connections

            parent_conn, child_conn = Pipe()
            worker = Process(
                target=_worker_process,
                args=(self._rulebase_dir, self._rule_ids_per_process[process_id], child_conn),
            )

            worker.start()
            parent_conn.send(self._rule_application_kwargs)

            self._processes[process_id] = worker
            self._connections[process_id] = parent_conn

    def _connection_try_call(
        self, process_id: int, restart_on_failure: bool, fn_name: str, args: tuple
    ) -> Any:
        MAX_RETRIES = 5
        for _ in range(MAX_RETRIES):
            try:
                return getattr(self._connections[process_id], fn_name)(*args)
            except (ConnectionResetError, BrokenPipeError):
                if restart_on_failure:
                    logger.warning(f"Pipe to process {process_id} seems to be corrupted")
                    self._restart_process(process_id)
                else:
                    break

    def _connection_try_send(self, process_id: int, restart_on_failure: bool, data: Any) -> None:
        self._connection_try_call(
            process_id=process_id,
            restart_on_failure=restart_on_failure,
            fn_name="send",
            args=(data,),
        )

    def _connection_try_recv(self, process_id: int, restart_on_failure: bool) -> Any:
        return self._connection_try_call(
            process_id=process_id, restart_on_failure=restart_on_failure, fn_name="recv", args=()
        )

    def _connection_try_close(self, process_id: int) -> None:
        self._connection_try_call(
            process_id=process_id, restart_on_failure=False, fn_name="close", args=()
        )

    def _close_processes(self, process_ids: list[int]) -> None:
        # Filter out processes that are already closed.
        process_ids = [process_id for process_id in process_ids if process_id in self._processes]

        for process_id in process_ids:
            self._connection_try_send(process_id=process_id, restart_on_failure=False, data=None)

        for process_id in process_ids:
            # First wait to see if the process finishes gracefully; if not, SIGTERM it.
            self._processes[process_id].join(timeout=2.0)
            self._processes[process_id].terminate()

        for process_id in process_ids:
            self._processes[process_id].join(timeout=2.0)

        for process_id in process_ids:
            try:
                self._processes[process_id].close()
            except ValueError as e:
                logger.warning(
                    f"Error trying to close a rule application process {process_id}: {e}"
                )

            self._connection_try_close(process_id=process_id)

            del self._processes[process_id]
            del self._connections[process_id]

    def _close_all_processes(self) -> None:
        self._close_processes(list(range(self._num_processes)))

    def _restart_process(self, process_id: int) -> None:
        logger.info(f"Restarting process {process_id}")

        self._close_processes([process_id])
        self._start_processes([process_id])
        self._connections[process_id].recv()  # Wait for the process to load its rulebase.

    def close(self) -> None:
        if not self._running:
            return

        self._close_all_processes()
        self._running = False

    def __getstate__(self) -> dict[str, Any]:
        """Pickling processes is not supported, we have to kill them before pickling the server."""
        self._close_all_processes()
        return self.__dict__

    def apply_rules(
        self,
        inputs: list[Molecule],
        rule_ids_to_apply: list[list[int]],
        timeout: Optional[float] = 60.0,
    ) -> list[list[list[RewriteResult]]]:
        """Apply the given rules to the given inputs in parallel.

        Args:
            inputs: Molecules to apply the rules to.
            rule_ids_to_apply: Rule IDs to apply for each input.
            timeout: Maximum amount of time to wait for applying a single rule. If we cannot obtain
                the result within that time (which could be due to the underlying `rdkit` code
                taking too long to match the rule or any other failures in the child process) then
                no results are returned for that specific rule, as if it did not match.

        Returns:
            A nested list of results in a layout matching `rule_ids_to_apply`.
        """
        # shape: `[batch_size, num_processes, num_rules_to_apply, 2]`
        rules_to_apply_per_process: list[list[Deque[tuple[int, int]]]] = [
            [deque() for _ in range(self._num_processes)] for _ in inputs
        ]

        for input_idx, rule_ids in enumerate(rule_ids_to_apply):
            for idx, rule_id in enumerate(rule_ids):
                rules_to_apply_per_process[input_idx][self._rule_id_to_process_id[rule_id]].append(
                    (idx, rule_id)
                )

        def send_tasks_to_process(process_id: int) -> None:
            for input, rules in zip(inputs, rules_to_apply_per_process):
                if rules[process_id]:
                    self._connection_try_send(
                        process_id=process_id,
                        restart_on_failure=True,
                        data=(input.smiles, [rule_id for _, rule_id in rules[process_id]]),
                    )

        for process_id in range(self._num_processes):
            send_tasks_to_process(process_id)

        # shape: `[batch_size, num_rules_to_apply, num_results]`
        batch_results: list[list[list[RewriteResult]]] = []
        for input_idx, input in enumerate(inputs):
            results: list[list[RewriteResult]] = [[] for _ in rule_ids_to_apply[input_idx]]

            for process_id in range(self._num_processes):
                while rules_to_apply_per_process[input_idx][process_id]:
                    idx, rule_id = rules_to_apply_per_process[input_idx][process_id].popleft()

                    if self._connections[process_id].poll(timeout):
                        results[idx] = self._connections[process_id].recv()
                    else:
                        logger.warning(
                            f"Timed out waiting for process {process_id} for {timeout} seconds "
                            f"(input: {input.smiles}, rule ID: {rule_id})"
                        )

                        self._restart_process(process_id)

                        # Send the remaining tasks to the restarted process (note that completed
                        # tasks were already popped from `rules_to_apply_per_process`).
                        send_tasks_to_process(process_id)

            batch_results.append(results)

        return batch_results



"""Script for preprocessing a raw dataset.

Running preprocessing requires knowing the model class that will be used during training; see
`train.py` for the currently available options.
"""

from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from omegaconf import MISSING
from syntheseus.reaction_prediction.data.reaction_sample import ReactionSample
from syntheseus.reaction_prediction.utils.config import get_config as cli_get_config
from syntheseus.reaction_prediction.utils.misc import cpu_count, set_random_seed

from retrochimera.cli.train import ModelClass, ModelTrainingConfig, build_model_from_config
from retrochimera.data.dataset import DataFold, DiskReactionDataset
from retrochimera.data.smiles_reaction_sample import SmilesReactionSample
from retrochimera.data.template_reaction_sample import TemplateReactionSample
from retrochimera.utils.logging import get_logger
from retrochimera.utils.training import preprocess_and_save

logger = get_logger(__name__)


@dataclass
class PreprocessConfig(ModelTrainingConfig):
    """Config for preprocessing a raw dataset."""

    data_dir: str = MISSING  # Directory containing raw data
    processed_data_dir: str = MISSING  # Directory for saving preprocessed data
    seed: int = 0  # Seed to use for all sources of randomness (Python, numpy, torch)
    max_num_samples: Optional[int] = None  # Maximum number of samples to truncate each fold to

    num_processes_preprocessing: int = cpu_count()
    rulebase_min_rule_support: Optional[int] = None
    rulebase_max_num_rules: Optional[int] = None


def main() -> None:
    config: PreprocessConfig = cli_get_config(argv=None, config_cls=PreprocessConfig)
    logger.info(f"Running preprocessing with the following config: {config}")

    set_random_seed(config.seed)

    processed_data_dir = Path(config.processed_data_dir)
    processed_data_dir.mkdir(parents=True, exist_ok=True)
    processed_data_path = processed_data_dir / "data.h5"

    if processed_data_path.exists():
        logger.info(f"Processed data file {processed_data_path} exists, nothing to do")
        return

    if config.rulebase_min_rule_support is None and config.rulebase_max_num_rules is None:
        logger.warning("Neither `rulebase_min_rule_support` nor `rulebase_max_num_rules` is set")

    reaction_sample_cls: type[ReactionSample]
    if config.model_class is ModelClass.SmilesTransformer:
        reaction_sample_cls = SmilesReactionSample
    else:
        reaction_sample_cls = TemplateReactionSample

    dataset = DiskReactionDataset(
        data_dir=config.data_dir,
        sample_cls=reaction_sample_cls,
        num_processes=config.num_processes_preprocessing,
        rulebase_min_rule_support=config.rulebase_min_rule_support,
        rulebase_max_num_rules=config.rulebase_max_num_rules,
    )

    logger.info(f"Saving the modified rulebase under {config.processed_data_dir}")
    dataset.rulebase.save_to_file(dir=processed_data_dir)

    model, _ = build_model_from_config(
        config=config, rulebase=dataset.rulebase, rulebase_dir=processed_data_dir
    )

    logger.info(
        f"Preprocessing datapoints: "
        f"{dataset.get_num_samples(DataFold.TRAIN)} train, "
        f"{dataset.get_num_samples(DataFold.VALIDATION)} validation, "
        f"{dataset.get_num_samples(DataFold.TEST)} test"
    )

    preprocess_and_save(
        save_path=processed_data_path,
        dataset=dataset,
        model=model,
        num_processes=config.num_processes_preprocessing,
        max_num_samples=config.max_num_samples,
    )


if __name__ == "__main__":
    main()



import abc
from pathlib import Path
from typing import Any, Generic, Optional, TypeVar, Union

import h5py
import pytorch_lightning as pl
from torch.utils.data import DataLoader, Dataset

from retrochimera.data.dataset import DataFold
from retrochimera.utils.misc import load_from_h5py

ProcessedSampleType = TypeVar("ProcessedSampleType")


class AbstractProcessedDataset(Dataset[ProcessedSampleType], Generic[ProcessedSampleType], abc.ABC):
    def __init__(self, h5_path: Union[str, Path], fold: DataFold):
        self.h5_path = h5_path
        self.fold = fold


class ProcessedDataset(AbstractProcessedDataset[ProcessedSampleType], Generic[ProcessedSampleType]):
    def __init__(self, h5_path: Union[str, Path], fold: DataFold):
        super().__init__(h5_path, fold)

        self._data = None

        with h5py.File(self.h5_path, mode="r", libver="latest") as f:
            self._len = len(f[self.fold.value])

    def __getitem__(self, idx: int) -> ProcessedSampleType:
        # Dataloader does not work with hdf5 when `num_workers > 1`. Solution below comes from
        # https://github.com/pytorch/pytorch/issues/11929.
        if self._data is None:
            self._data = h5py.File(self.h5_path, mode="r", libver="latest")[self.fold.value]

        assert self._data is not None
        return load_from_h5py(self._data[str(idx)])

    def __len__(self) -> int:
        return self._len


class ProcessedDataModule(pl.LightningDataModule):
    def __init__(
        self,
        h5_path: Union[str, Path],
        data_loader_kwargs: Optional[dict[str, Any]] = None,
        processed_dataset_cls: type[AbstractProcessedDataset] = ProcessedDataset,
    ):
        super().__init__()
        self.prepare_data_per_node = False

        self._h5_path = h5_path
        self._data_loader_kwargs = data_loader_kwargs or {}
        self.processed_dataset_cls = processed_dataset_cls

    def _dataloader(self, fold: DataFold, shuffle: bool) -> DataLoader:
        kwargs = {"shuffle": shuffle} | self._data_loader_kwargs
        return DataLoader(
            dataset=self.processed_dataset_cls(h5_path=self._h5_path, fold=fold), **kwargs
        )

    def train_dataloader(self) -> DataLoader:
        return self._dataloader(fold=DataFold.TRAIN, shuffle=True)

    def val_dataloader(self) -> DataLoader:
        return self._dataloader(fold=DataFold.VALIDATION, shuffle=False)

    def test_dataloader(self) -> DataLoader:
        return self._dataloader(fold=DataFold.TEST, shuffle=False)



from typing import Any

from pytorch_lightning import callbacks as pl_callbacks
from pytorch_lightning.callbacks import Callback

from retrochimera.utils.logging import get_logger

logger = get_logger(__name__)


class OptLRMonitor(Callback):
    def __init__(self):
        super().__init__()

    def on_train_batch_start(self, trainer, *args, **kwargs):
        # Only support one optimizer
        opt = trainer.optimizers[0]

        # Only support one param group
        stats = {"lr-Adam": opt.param_groups[0]["lr"]}
        trainer.logger.log_metrics(stats, step=trainer.global_step)


class ModelCheckpoint(pl_callbacks.ModelCheckpoint):
    """Custom `ModelCheckpoint` callback which fixes checkpoint paths upon directory change.

    Note that we keep the original class name for backward compatibility.
    """

    def load_state_dict(self, state_dict: dict[str, Any]) -> None:
        dirpath_from_ckpt = state_dict.get("dirpath", self.dirpath)

        if self.dirpath != dirpath_from_ckpt:
            logger.warning(
                f"Current checkpoint directory {self.dirpath} differs from {dirpath_from_ckpt} "
                "found in the checkpoint. Updating all paths with the new prefix."
            )
            assert self.dirpath is not None

            for key in [
                "dirpath",
                "best_model_path",
                "last_model_path",
                "best_k_models",
                "kth_best_model_path",
            ]:
                value = state_dict[key]

                if not value:
                    continue

                if isinstance(value, str):
                    assert value.startswith(dirpath_from_ckpt)
                    new_value = self.dirpath + value.removeprefix(dirpath_from_ckpt)
                else:
                    assert isinstance(value, dict)
                    assert all(k.startswith(dirpath_from_ckpt) for k in value)
                    new_value = {
                        self.dirpath + k.removeprefix(dirpath_from_ckpt): v
                        for k, v in value.items()
                    }

                logger.info(f"Updated {key} from {value} to {new_value}")
                state_dict[key] = new_value

        super().load_state_dict(state_dict)



import argparse
import math
import multiprocessing
from abc import abstractmethod
from typing import Any, Generic, Sequence, TypeVar

import torch
from syntheseus import Molecule, Reaction, SingleProductReaction
from syntheseus.interface.reaction import ReactionMetaData
from syntheseus.reaction_prediction.inference_base import ExternalBackwardReactionModel
from syntheseus.reaction_prediction.utils.inference import (
    get_unique_file_in_dir,
    process_raw_smiles_outputs_backwards,
)
from syntheseus.reaction_prediction.utils.misc import suppress_outputs

from retrochimera.models.smiles_transformer import SmilesTransformerModel as TransformerModel
from retrochimera.opennmt.decode.translator import Translator
from retrochimera.utils.logging import get_logger
from retrochimera.utils.root_aligned import get_product_roots

logger = get_logger(__name__)


InputType = TypeVar("InputType")
ReactionType = TypeVar("ReactionType", bound=Reaction)


class AbstractSmilesTransformerModel(Generic[InputType, ReactionType]):
    def __init__(
        self,
        *args,
        beam_size: int = 20,
        augmentation_size: int = 10,
        max_generated_seq_len: int = 512,
        probability_from_score_temperature: float = 3.0,
        filter_duplicate_augmentations: bool = True,
        **kwargs,
    ) -> None:
        """Initializes the SmilesTransformer model wrapper.

        This base class is shared between the backward and forward model variants.

        Assumed format of the model directory:
        - `model_dir` contains the model checkpoint as the only `*.ckpt` file
        - `model_dir` contains the vocab path as the only `*.txt` file
        """

        super().__init__(*args, **kwargs)

        if not (hasattr(self, "model_dir") and hasattr(self, "device")):
            raise ValueError(
                "Class based on `AbstractSmilesTransformerModel` should extended `ReactionModel`"
            )

        # There should be exaclty one `*.ckpt` file under `model_dir`.
        chkpt_path = get_unique_file_in_dir(self.model_dir, pattern="*.ckpt")
        vocab_path = get_unique_file_in_dir(self.model_dir, pattern="*.txt")

        with suppress_outputs():
            self.model = TransformerModel.load_from_checkpoint(chkpt_path, vocab_path=vocab_path)

        self.model = self.model.to(self.device)
        self.model.eval()

        # Configure the inference parameters.
        self.beam_size = beam_size
        self.augmentation_size = augmentation_size
        self.max_generated_seq_len = max_generated_seq_len
        assert (
            self.max_generated_seq_len <= self.model.max_seq_len
        ), f"The maximum generated sequence length should be less than or equal to the maximum sequence length supported by the model (e.g., for position encoding): {self.model.max_seq_len}."

        self.probability_from_score_temperature = probability_from_score_temperature
        self.filter_duplicate_augmentations = filter_duplicate_augmentations

        self.model.tokenizer.log_vocab_info()
        logger.info(f"Beam size: {self.beam_size}")
        logger.info(f"Augmentation size: {self.augmentation_size}")
        logger.info(f"Maximum generated sequence length: {self.max_generated_seq_len}")
        logger.info(f"Filter duplicate augmentations: {self.filter_duplicate_augmentations}")

    def get_parameters(self):
        return self.model.parameters()

    def _build_kwargs_from_scores(
        self, scores: list[float], additional_info: dict
    ) -> list[ReactionMetaData]:
        from syntheseus.reaction_prediction.inference.root_aligned import RootAlignedModel

        kwargs_list = RootAlignedModel.build_prediction_kwargs_from_scores(
            scores,
            num_augmentations=self.augmentation_size,
            beam_size=self.beam_size,
            probability_from_score_temperature=self.probability_from_score_temperature,
        )

        position_prob_tuples = additional_info["position_prob_tuples"]
        adaptive_augmentation_size = additional_info["adaptive_augmentation_size"]
        effective_augmentation_size = additional_info["effective_augmentation_size"]

        assert len(scores) == len(
            position_prob_tuples
        ), f"Number of scores ({len(scores)}) does not match number of info entries ({len(position_prob_tuples)})"

        for kwargs, position_prob_tuple in zip(kwargs_list, position_prob_tuples):
            kwargs.update(
                {  # type: ignore[misc]
                    "position_prob_tuple": position_prob_tuple,
                    "adaptive_augmentation_size": adaptive_augmentation_size,
                    "effective_augmentation_size": effective_augmentation_size,
                }
            )

        return kwargs_list

    def _smiles_to_batch(self, smiles: list[str]) -> dict[str, Any]:
        token_ids, padding_mask = self.model.tokenizer.encode(smiles, pad=True, right_padding=True)

        # Convert inputs to the model to tensors.
        return {
            "encoder_input": torch.tensor(token_ids),  # shape: (batch_size, src_len)
            "encoder_padding_mask": torch.tensor(
                padding_mask, dtype=torch.bool
            ),  # shape: (batch_size, src_len)
        }

    @abstractmethod
    def _augment_input(self, input: InputType) -> list[str]:
        pass

    @abstractmethod
    def _process_raw_smiles_outputs(
        self, input: InputType, output_list: list[str], metadata_list: list[ReactionMetaData]
    ) -> Sequence[ReactionType]:
        pass

    def _get_reactions(
        self, inputs: list[InputType], num_results: int
    ) -> list[Sequence[ReactionType]]:
        from retrochimera.utils.root_aligned_score import (
            canonicalize_smiles_clear_map,
            compute_rank,
        )

        # Step 1: Perform data augmentation on the input side (and convert to SMILES along the way).
        augmented_inputs: list[str] = []

        # Set up `opt` for the `compute_rank` function, originally from `root_aligned.score`.
        opt = argparse.Namespace()
        setattr(opt, "synthon", False)
        setattr(opt, "beam_size", self.beam_size)

        adaptive_augmentation_sizes = []
        for input in inputs:
            augmented_input = self._augment_input(input)

            if self.filter_duplicate_augmentations:
                augmented_input = list(dict.fromkeys(augmented_input))

            augmented_inputs.extend(augmented_input)
            adaptive_augmentation_sizes.append(len(augmented_input))

        # Step 2: Map from `InputType` class to `torch.Tensor`s.
        augmented_batch = self._smiles_to_batch(augmented_inputs)

        augmented_batch_to_device = {
            key: val.to(self.device) if type(val) == torch.Tensor else val  # type: ignore[attr-defined]
            for key, val in augmented_batch.items()
        }

        # We have to set `num_beams` as an attribute of the model.
        self.model.num_beams = self.beam_size
        translator = Translator(
            model=self.model,
            tokenizer=self.model.tokenizer,
            n_best=self.beam_size,
            max_length=self.max_generated_seq_len,
            beam_size=self.beam_size,
        )

        batch = {}
        src = (
            augmented_batch_to_device["encoder_input"].transpose(0, 1).contiguous().unsqueeze(2),
            torch.sum(~augmented_batch_to_device["encoder_padding_mask"], dim=1),
        )
        batch_size = src[0].size(1)
        assert src[0].size(1) == src[1].size(0)
        batch["src"] = src  # tuple[Tensor, Tensor]: (padded_src_len, batch_size, 1), (batch_size,)
        batch["batch_size"] = batch_size

        translate_results = translator.translate_batch(batch, attn_debug=False)
        augmented_batch_output_token_ids = translate_results[
            "predictions"
        ]  # list[list[LongTensor]]: For each batch, holds a list of beam prediction sequences
        augmented_batch_smiles = augmented_batch_output_token_ids

        # Obtain probabilities of the predictions.
        augmented_batch_log_probs = translate_results["scores"]  # list[list[float]]
        augmented_batch_scores = [
            [math.exp(log_prob) for log_prob in log_probs]
            for log_probs in augmented_batch_log_probs
        ]  # shape: (input_size * adaptive_augmentation_size, num_beams)

        lines = []  # shape: (input_size * adaptive_augmentation_size * num_beams)
        for i in range(len(augmented_batch_smiles)):
            for j in range(len(augmented_batch_smiles[i])):
                line = self.model.tokenizer.decode([augmented_batch_smiles[i][j].tolist()])
                assert isinstance(line, list)
                assert len(line) == 1
                assert isinstance(line[0], str)
                lines.append((line[0], augmented_batch_scores[i][j]))

        raw_predictions = []
        pool = multiprocessing.Pool(4)

        raw_predictions = pool.map(
            func=canonicalize_smiles_clear_map, iterable=lines
        )  # canonicalize reactants and modify illegal reactants into empty strings
        pool.close()
        pool.join()

        predictions = []
        left_index = 0
        for augmentation_size in adaptive_augmentation_sizes:
            curr_input_predictions = []
            for i in range(augmentation_size):
                curr_input_predictions.append(
                    raw_predictions[left_index : left_index + self.beam_size]
                )
                left_index += self.beam_size
            predictions.append(curr_input_predictions)

        # Rank legal reactants from all augmentations and beams.
        ranked_results = []
        ranked_scores = []
        ranked_additional_info = []

        for i in range(len(predictions)):
            (
                rank,
                position_prob_info,
                adaptive_augmentation_size,
                effective_augmentation_size,
            ) = compute_rank(predictions[i], opt=opt)
            # `adaptive_augmentation_size` is the number of augmentations performed on the input product molecule.
            # `effective_augmentation_size` is the number of augmentations where the transformer model returns at least one valid SMILES prediction in the beam.
            # `rank` is a dictionary with (reactant_smiles, max_frag_smiles) as keys and their computed scores for ranking as values.
            rank_list = list(zip(rank.keys(), rank.values()))
            rank_list.sort(key=lambda x: x[1], reverse=True)
            rank_list = rank_list[:num_results]  # truncate to `num_results` results
            ranked_results.append([item[0][0] for item in rank_list])  # output reactant SMILES
            ranked_scores.append([item[1] for item in rank_list])  # output scores used for ranking
            ranked_additional_info.append(
                {
                    "position_prob_tuples": [position_prob_info[item[0]] for item in rank_list],
                    "adaptive_augmentation_size": adaptive_augmentation_size,
                    "effective_augmentation_size": effective_augmentation_size,
                }
            )

        return [
            self._process_raw_smiles_outputs(
                input, outputs, self._build_kwargs_from_scores(scores, additional_info)
            )
            for input, outputs, scores, additional_info in zip(
                inputs, ranked_results, ranked_scores, ranked_additional_info
            )
        ]

    def compute_probs(
        self, reaction_smiles: list[str], minibatch_size: int = 32
    ) -> tuple[list[float], list[float]]:
        """Compute total and average probabilities for a list of reaction SMILES strings."""
        return self.model.compute_probs(reaction_smiles, minibatch_size=minibatch_size)


class SmilesTransformerModel(
    AbstractSmilesTransformerModel[Molecule, SingleProductReaction], ExternalBackwardReactionModel
):
    def _augment_input(self, input: Molecule) -> list[str]:
        from root_aligned.preprocessing.generate_PtoR_data import clear_map_canonical_smiles

        augmented_input = []

        product_roots = get_product_roots(
            product_atom_ids=[i + 1 for i in range(input.rdkit_mol.GetNumAtoms())],
            num_augmentations=self.augmentation_size,
        )

        for pro_root_atom_id in product_roots:
            pro_root = pro_root_atom_id - 1
            if pro_root_atom_id <= 0:
                pro_root = -1

            augmented_input.append(
                clear_map_canonical_smiles(input.smiles, canonical=True, root=pro_root)
            )

        return augmented_input

    def _process_raw_smiles_outputs(
        self, input: Molecule, output_list: list[str], metadata_list: list[ReactionMetaData]
    ) -> Sequence[SingleProductReaction]:
        return process_raw_smiles_outputs_backwards(input, output_list, metadata_list)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

from syntheseus.interface.molecule import SMILES_SEPARATOR
from syntheseus.reaction_prediction.data.reaction_sample import ReactionType

from retrochimera.data.template_reaction_sample import TemplateReactionSample


@dataclass(frozen=True)
class SmilesReactionSample(TemplateReactionSample):
    """Reaction sample that also loads the SMILES with no canonicalization or standardization."""

    # We need to provide default values due to inheritance, but never expect these to be kept.
    raw_products_smiles: str = field(default="", hash=True, compare=True)
    raw_reactants_smiles: str = field(default="", hash=True, compare=True)

    @classmethod
    def from_dict(cls: type[ReactionType], data: dict[str, Any]) -> ReactionType:
        for key in ["reactants", "products"]:
            raw_key = f"raw_{key}_smiles"
            if raw_key not in data:
                data[raw_key] = SMILES_SEPARATOR.join([d["smiles"] for d in data[key]])

        return super().from_dict(data)



from pathlib import Path

from syntheseus import Molecule
from syntheseus.reaction_prediction.data.reaction_sample import ReactionSample

from retrochimera.data.processing import (
    AtomMappingProcessingStep,
    NumAtomsProcessingStep,
    NumReactantsProcessingStep,
    OneMainProductProcessingStep,
    ProcessingStep,
    ProductAmongReactantsProcessingStep,
)


def make_sample(reaction_smiles: str) -> ReactionSample:
    sample = ReactionSample.from_reaction_smiles_strict(reaction_smiles, mapped=True)
    sample.metadata["original_str"] = reaction_smiles  # type: ignore[typeddict-item]

    return sample


def run_step_and_compare(
    step: ProcessingStep, inputs: list[ReactionSample], expected_outputs: list[ReactionSample]
) -> None:
    outputs = list(step.process_samples(inputs))

    for output, expected_output in zip(outputs, expected_outputs):
        assert output.reactants == expected_output.reactants
        assert output.products == expected_output.products
        assert (
            output.metadata.get("repaired_str", output.metadata["original_str"])  # type: ignore[typeddict-item]
            == expected_output.metadata["original_str"]  # type: ignore[typeddict-item]
        )


def test_num_reactants_processing_step(tmpdir: Path) -> None:
    samples = [
        make_sample("[C:1]>>[C:1]"),
        make_sample("[C:1].[C:2]>>[C:1][C:2]"),
        make_sample("[C:1].[C:2].[C:3]>>[C:1][C:2][C:3]"),
    ]

    run_step_and_compare(
        step=NumReactantsProcessingStep(name="1", output_dir=tmpdir, max_reactants_num=2),
        inputs=samples,
        expected_outputs=samples[:2],
    )


def test_one_main_product_processing_step(tmpdir: Path) -> None:
    samples = [
        make_sample("[C:1]>>[C:1]"),
        make_sample("[C:1].[C:2]>>[C:1][C:2]"),
        make_sample("[C:1].[C:2]>>[C:1].[C:2]"),
        make_sample("[C:1].[C:2].[C:3]>>[C:1].[C:2][C:3]"),
        make_sample("[C:1].[C:2].[C:3]>>[C:1][C:2][C:3]"),
        make_sample("[C:1].[C:2].[C:3].[C:4]>>[C:1][C:2][C:3].[C:4]"),
        make_sample("[C:1].[C:2].[C:3].[C:4].[C:5]>>[C:1][C:2][C:3].[C:4][C:5]"),
    ]

    run_step_and_compare(
        step=OneMainProductProcessingStep(
            name="1",
            output_dir=tmpdir,
            min_product_atoms=2,
            common_products={Molecule("[C][C][C]")},
        ),
        inputs=samples,
        expected_outputs=[
            samples[1],
            make_sample("[C:1].[C:2].[C:3]>>[C:2][C:3]"),  # `samples[3]` repaired
            make_sample("[C:1].[C:2].[C:3].[C:4].[C:5]>>[C:4][C:5]"),  # `samples[6]` repaired
        ],
    )


def test_num_atoms_processing_step(tmpdir: Path) -> None:
    samples = [
        make_sample("[C:1].[C:2]>>[C:1][C:2]"),
        make_sample("[C:1].[C:2].[C:3]>>[C:1][C:2][C:3]"),
        make_sample("[C:1].[C:2].CN>>[C:1][C:2]"),
        make_sample("[C:1].[C:2].CN=O>>[C:1][C:2]"),
    ]

    run_step_and_compare(
        step=NumAtomsProcessingStep(
            name="1", output_dir=tmpdir, max_product_atoms=2, max_reactants_to_product_ratio=2
        ),
        inputs=samples,
        expected_outputs=[samples[0], samples[2]],
    )


def test_product_among_reactants_processing_step(tmpdir: Path) -> None:
    samples = [
        make_sample("[C:1].[C:2]>>[C:1][C:2]"),
        make_sample("[C:1][C:2]>>[C:1][C:2]"),
        make_sample("[C:1].[C:2].[C][C]>>[C:1][C:2]"),
    ]

    run_step_and_compare(
        step=ProductAmongReactantsProcessingStep(name="1", output_dir=tmpdir),
        inputs=samples,
        expected_outputs=samples[:1],
    )


def test_atom_mapping_processing_step(tmpdir: Path) -> None:
    samples = [
        make_sample("[C:1].[C:2]>>[C:1][C:2]"),
        make_sample("[C:1].[C:2].[C:3]>>[C:1][C:2]"),
        make_sample("[C:1].[C:2]>>[C:1][C:2][C:3]"),
        make_sample("CC>>CC"),
        make_sample("NC=O.[C:1]>>[C:1]"),
        make_sample("[C:1].[C:1]>>[C:1][C:2]"),
        make_sample("[C:1].[C:2]>>[C:1][C:1]"),
        make_sample("[C:1].[C:2]>>[C:1][C:3]"),
    ]

    run_step_and_compare(
        step=AtomMappingProcessingStep(name="1", output_dir=tmpdir),
        inputs=samples,
        expected_outputs=[
            samples[0],
            make_sample("[C:1].[C:2]>>[C:1][C:2]"),  # `samples[1]` repaired
            make_sample("[C:1].[C:2]>>[C][C:2][C:1]"),  # `samples[2]` repaired
            make_sample("[C:1]>>[C:1]"),  # `samples[4]` repaired
            make_sample("[C:1]>>[C][C:1]"),  # `samples[7]` repaired
        ],
    )



"""Fingerprint methods to featurize molecules and reactions."""

import copy
from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum
from functools import partial
from typing import Iterator, Union

import joblib
import numpy as np
from numpy.typing import NDArray
from rdkit import Chem
from rdkit.Chem import AllChem, rdFingerprintGenerator, rdmolops
from rdkit.Chem.rdChemReactions import (
    CreateDifferenceFingerprintForReaction,
    FingerprintType,
    ReactionFingerprintParams,
)
from scipy import sparse as sp
from syntheseus.interface.molecule import Molecule


def fp_to_dense_np(fp, size: int = 16384) -> NDArray:
    """Convert rdkit counted fingerprint to dense numpy array, with modulo-folding to given size.

    Args:
        fp: RDKit count fingerprint.
        size: Length to fold the fingerprint to (via mod).

    Returns:
        Fingerprint as a dense `numpy` array.
    """

    array = np.zeros(size)

    for i, j in fp.GetNonzeroElements().items():
        k = i % size
        array[k] += j

    return array


def fp_to_sparse_array(fp, size: int = 2048) -> sp.dok_matrix:
    """Converts `rdkit` count fingerprint to `scipy` sparse array.

    The fingerprint is modulo-folded to the given size. The returned array, which is in
    `dok_matrix` form, should be converted into the needed array later on, e.g. CSR matrix.

    Args:
        fp: RDKit count fingerprint.
        size: Length to fold the fingerprint to (via mod).

    Returns:
        Fingerprint as a sparse `scipy.sparse.dok_matrix` array.
    """
    array = sp.dok_matrix((1, size), dtype="i")

    for i, j in fp.GetNonzeroElements().items():
        k = i % size
        array[0, k] += j

    return array


def tanimoto(fp1: NDArray, fp2: NDArray) -> NDArray:
    assert fp1.shape == fp2.shape
    assert len(fp1.shape) == 2

    ab = (fp1 * fp2).sum(axis=-1)
    a = (fp1 * fp1).sum(axis=-1)
    b = (fp2 * fp2).sum(axis=-1)

    c = a + b - ab

    return np.divide(a, c, out=np.zeros_like(ab), where=c != 0)


def tanimoto_sparse(fp1, fp2):
    assert fp1.shape == fp2.shape
    assert len(fp1.shape) == 2

    ab = np.asarray(fp1.multiply(fp2).sum(axis=-1)).flatten()
    a = np.asarray(fp1.multiply(fp1).sum(axis=-1)).flatten()
    b = np.asarray(fp2.multiply(fp2).sum(axis=-1)).flatten()

    return ab / (a + b - ab)


def calc_morgan(mol: Chem.Mol, radius: int = 2):
    Chem.rdmolops.GetSSSR(mol)
    return AllChem.GetMorganFingerprint(mol, radius)


def calc_ap(mol: Chem.Mol, max_path: int = 3, n_bits: int = 1001107):
    return AllChem.GetHashedAtomPairFingerprint(mol, maxLength=max_path, nBits=n_bits)


def calc_rdkfp(mol: Chem.Mol, max_path: int = 8, branched: bool = True):
    return rdmolops.UnfoldedRDKFingerprintCountBased(mol, maxPath=max_path, branchedPaths=branched)


def calc_count_fp(
    mol: Chem.Mol, size: int = 32768, radius: int = 2, fold_mod: int = 32749
) -> NDArray:
    assert fold_mod <= size

    fpgen = rdFingerprintGenerator.GetMorganGenerator(radius=radius)

    fingerprint = np.zeros((size,), dtype=np.float32)
    for index, count in fpgen.GetSparseCountFingerprint(mol).GetNonzeroElements().items():
        fingerprint[index % fold_mod] += count

    return fingerprint


def calc_rxn_morgan(
    rxn: AllChem.ChemicalReaction, size: int = 4096, fp_type=FingerprintType.MorganFP
):
    params = ReactionFingerprintParams(False, 0, 1, 0, size, fp_type)
    return CreateDifferenceFingerprintForReaction(rxn, params)


def calc_rxn_fp(rxn: AllChem.ChemicalReaction, fp_fn, size: int = 8192):
    output = sp.dok_matrix((1, size), dtype="i")

    for mol in rxn.GetProducts():
        mol.UpdatePropertyCache(False)
        fp = fp_fn(mol)
        output += fp_to_sparse_array(fp, size=size)

    for mol in rxn.GetReactants():
        mol.UpdatePropertyCache(False)
        fp = fp_fn(mol)
        output -= fp_to_sparse_array(fp, size=size)

    return output.tocsr()


def sum_fps(fps):
    summed_fp = copy.deepcopy(fps[0])
    for fp in fps[1:]:
        summed_fp += fp
    return summed_fp


@dataclass
class ChemData:
    embedding: Union[NDArray, sp.dok_matrix]
    valid: bool


class ChemicalEmbedder:
    """Abstract class to featurize chemical entities (molecules and reactions) in batch."""

    def __init__(self, size: int = 1024, sparse: bool = True):
        self.size = size
        self.sparse = sparse

    def embed_list(self, smiles_list: list[str], n_jobs: int = -1) -> Iterator[ChemData]:
        pool = joblib.Parallel(n_jobs=n_jobs)
        jobs = (joblib.delayed(self._convert)(s) for s in smiles_list)

        return iter(pool(jobs))

    def embed(self, smiles_list: list[str], n_jobs: int = -1) -> tuple[NDArray, NDArray]:
        """Featurize a list of molecules or reactions.

        Args:
            smiles_list: Chemical entities to featurize (molecules or reactions).
            n_jobs: Number of parallel processes to use.

        Returns:
            Two arrays of the same length as `smiles_list`. The first contains the feature vectors,
            the second is a binary mask which indicates whether the featurization worked.
        """
        embeddings = []
        mask = []

        for result in self.embed_list(smiles_list, n_jobs):
            embeddings.append(result.embedding)
            mask.append(result.valid)

        return self._postprocess(embeddings, mask)

    def _postprocess(self, X, mask):
        if self.sparse:
            return self._post_proc_sparse(sp.vstack(X, format="csr")), np.array(mask)
        else:
            return self._post_proc_dense(np.vstack(X)), np.array(mask)

    def _post_proc_sparse(self, X):
        return X

    def _post_proc_dense(self, X):
        return X

    @abstractmethod
    def _convert(self, smiles):
        pass


class MoleculeEmbedder(ChemicalEmbedder):
    """Class to featurize molecules (fingerprint hardcoded as Morgan2 / ECFP4)."""

    def __init__(self, size: int = 1024, sparse: bool = True, log_x: bool = True) -> None:
        super().__init__(size, sparse)
        self.log_x = log_x

    def _convert(self, smiles: str) -> ChemData:
        mol = Chem.MolFromSmiles(smiles)

        if mol:
            fp = calc_morgan(mol, 2)
            return self._return_fp(fp)
        else:
            return ChemData(embedding=np.zeros(self.size), valid=False)

    def convert_mol(self, mol: Molecule) -> ChemData:
        return self._convert(mol.smiles)

    def _return_fp(self, fp) -> ChemData:
        if self.sparse:
            return ChemData(embedding=fp_to_sparse_array(fp, self.size), valid=True)
        else:
            return ChemData(embedding=fp_to_dense_np(fp, self.size), valid=True)

    def _post_proc_dense(self, X):
        if self.log_x:
            return np.log(X + 1)
        else:
            return X

    def _post_proc_sparse(self, X):
        if self.log_x:
            X.data = np.log(X.data + 1)
            return X
        else:
            return X


class FPType(Enum):
    ecfp = 1
    apfp = 2
    pathfp = 3


def get_fp_function(fptype: FPType, max_path: int = 3, radius: int = 2, branched: bool = True):
    assert fptype in FPType

    if fptype == FPType.ecfp:
        return partial(calc_morgan, radius=radius)
    if fptype == FPType.apfp:
        return partial(calc_rdkfp, max_path=max_path, branched=branched)
    if fptype == FPType.pathfp:
        return partial(calc_ap, max_path=max_path)



from typing import Callable, Optional

import torch
from syntheseus.interface.molecule import Molecule
from torch import Tensor

from retrochimera.chem.descriptors import ChemData, MoleculeEmbedder
from retrochimera.encoders.base import Encoder, EncoderOutput

BatchType = tuple[Tensor, Tensor]


class ChemEncoder(Encoder[Molecule, ChemData, BatchType]):
    """Implements a molecule encoder based on fixed chemical features such as fingerprints."""

    def __init__(self, fingerprint_dim: int, sparse: bool = True, log_x: bool = True) -> None:
        super().__init__()

        self.fingerprint_dim = fingerprint_dim
        self.embedder = MoleculeEmbedder(size=fingerprint_dim, sparse=sparse, log_x=log_x)

    @property
    def preprocess(self) -> Callable[[Molecule], ChemData]:
        return self.embedder.convert_mol

    def collate(self, inputs: list[ChemData]) -> BatchType:
        embeddings, masks = self.embedder._postprocess(
            X=[input.embedding for input in inputs], mask=[input.valid for input in inputs]
        )
        return (
            torch.as_tensor(embeddings.todense(), dtype=torch.float),
            torch.as_tensor(masks.flatten()),
        )

    def forward(self, batch: BatchType) -> EncoderOutput:
        mol_outputs, _ = batch  # The second element is a validity mask, which we ignore for now.
        return EncoderOutput(atom_outputs=None, bond_outputs=None, mol_outputs=mol_outputs)

    def move_batch_to_device(self, batch: BatchType, device: str) -> BatchType:
        outputs, validity_mask = batch
        return outputs.to(device), validity_mask.to(device)

    @property
    def atom_out_channels(self) -> Optional[int]:
        return None

    @property
    def bond_out_channels(self) -> Optional[int]:
        return None

    @property
    def mol_out_channels(self) -> int:
        return self.fingerprint_dim



"""Script for running initial processing of a dataset and splitting it into folds.

The processing hyperparameters can be overriden, but by default they map to the following settings:
1. Remove reactions with more than 4 reactants.
2. Remove reactions with zero or several main products (defined as those having at least 5 atoms and
   not being among "common products", i.e. products appearing in the data at least 1000 times).
3. Remove reactions with product having more than 100 atoms or with ratio between number of reactant
   and product atoms being more than 20.
4. Remove reactions where the main product is also present among the reactants.
5. Refine reactions by removing atom mapping numbers appearing only on one side, and dropping
   reactants with no mapped atoms. Remove reactions with an invalid mapping or no mapping left.

Split into folds is random after grouping by product. Optionally, a `dataset_to_follow_dir` argument
allows for splitting in accordance with an external (already split) dataset.
"""

import random
from collections import Counter, defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, Optional

from syntheseus.interface.bag import Bag
from syntheseus.interface.molecule import Molecule
from syntheseus.reaction_prediction.data.reaction_sample import ReactionSample
from syntheseus.reaction_prediction.utils.config import get_config as cli_get_config
from syntheseus.reaction_prediction.utils.misc import cpu_count, set_random_seed
from tqdm import tqdm

from retrochimera.data.dataset import DataFold
from retrochimera.data.processing import (
    AtomMappingProcessingStep,
    NumAtomsProcessingStep,
    NumReactantsProcessingStep,
    OneMainProductProcessingStep,
    ProductAmongReactantsProcessingStep,
    parse_samples,
)
from retrochimera.data.utils import load_raw_reactions_file, load_raw_reactions_files
from retrochimera.utils.logging import get_logger
from retrochimera.utils.misc import silence_rdkit_warnings

logger = get_logger(__name__)


@dataclass
class SplitDatasetConfig:
    """Config for splitting a given dataset."""

    input_path: str  # path to the dataset to be split (either Pistachio or USPTO format)
    output_dir: str  # output directory for saving the resulting folds
    num_processes: int = cpu_count()  # number of processes to use for processing the SMILES

    # Directory containing the folds of a pre-split datset. These are used to define products that
    # should preferrably be placed into the train/validation/test set. This setting is useful if one
    # wants to train on one dataset and validate on another pre-split one while minimizing leakage.
    dataset_to_follow_dir: Optional[str] = None

    # Fraction of samples to place into each fold.
    val_frac: float = 0.05
    test_frac: float = 0.05

    # Hyperparamters for filtering some bad samples.
    max_reactants_num: int = 4
    min_product_atoms: int = 5
    max_product_occurrences: int = 1000
    max_product_atoms: int = 100
    max_reactants_to_product_ratio: int = 20


def filter_dataset(data: list[ReactionSample], config: SplitDatasetConfig) -> list[ReactionSample]:
    final_data = []

    changed_samples_dir = Path(config.output_dir) / "changed_samples"
    changed_samples_dir.mkdir(parents=True, exist_ok=True)

    # Step 1: Drop samples with too many reactants.
    data = list(
        NumReactantsProcessingStep(
            name="1", output_dir=changed_samples_dir, max_reactants_num=config.max_reactants_num
        ).process_samples(data)
    )

    # Create and save occurrence count of product molecules.
    common_products: Counter = Counter()
    for sample in data:
        common_products.update(sample.products)

    common_products = Counter(
        {p: cnt for p, cnt in common_products.items() if cnt >= config.max_product_occurrences}
    )

    with open(changed_samples_dir / "common_products.txt", "w") as f:
        for product, count in common_products.most_common():
            f.write(f"{product.smiles} {count}\n")

    # Step 2: Drop samples with more than one main product.
    # A non-main product either has fewer than `min_product_atoms` atoms or occurs at least
    # `max_product_occurrences` times (over the entire dataset).
    one_main_product_step = OneMainProductProcessingStep(
        name="2",
        output_dir=changed_samples_dir,
        min_product_atoms=config.min_product_atoms,
        common_products=set(common_products),
    )

    # Step 3: Drop samples with large product or large ratio between reactants and product.
    num_atoms_step = NumAtomsProcessingStep(
        name="3",
        output_dir=changed_samples_dir,
        max_product_atoms=config.max_product_atoms,
        max_reactants_to_product_ratio=config.max_reactants_to_product_ratio,
    )

    # Step 4: Drop samples where the main product is also present among the reactants.
    product_among_reactants_step = ProductAmongReactantsProcessingStep(
        name="4", output_dir=changed_samples_dir
    )

    # Step 5: Fix samples with minor mapping issues and drop those that cannot be fixed.
    atom_mapping_step = AtomMappingProcessingStep(name="5", output_dir=changed_samples_dir)

    data_iterable: Iterable[ReactionSample] = data
    for step in [
        one_main_product_step,
        num_atoms_step,
        product_among_reactants_step,
        atom_mapping_step,
    ]:
        data_iterable = step.process_samples(data_iterable)

    for sample in tqdm(data_iterable, total=len(data)):
        # Clear rdkit molecules to save memory.
        for product in sample.products:
            if "rdkit_mol" in product.metadata:
                del product.metadata["rdkit_mol"]
        for reactant in sample.reactants:
            if "rdkit_mol" in reactant.metadata:
                del reactant.metadata["rdkit_mol"]

        if "repaired_str" in sample.metadata:
            sample.metadata["original_str"] = sample.metadata["repaired_str"]  # type: ignore[typeddict-item]

        final_data.append(sample)

    return final_data


def split_dataset(config: SplitDatasetConfig) -> None:
    raw_data = load_raw_reactions_file(config.input_path)
    logger.info(f"Read {len(raw_data)} raw samples to split")

    data = parse_samples(raw_data, num_processes=config.num_processes)
    logger.info(f"Left with {len(data)} valid samples\n")

    data = filter_dataset(data, config)
    logger.info(f"Left with {len(data)} samples after filtering")

    fold_target_size: dict[DataFold, int] = {
        DataFold.VALIDATION: int(config.val_frac * len(data)),
        DataFold.TEST: int(config.test_frac * len(data)),
    }
    fold_target_size[DataFold.TRAIN] = len(data) - sum(fold_target_size.values())

    fold_sizes_joined = "\n".join([f"{fold.value}: {fold_target_size[fold]}" for fold in DataFold])
    logger.info(f"Fold target sizes:\n{fold_sizes_joined}")

    assert min(fold_target_size.values()) >= 0, "Target fold sizes must be positive"

    data_grouped: dict[Bag[Molecule], list[ReactionSample]] = defaultdict(list)
    num_products_count: dict[int, int] = defaultdict(int)
    num_products_example: dict[int, str] = {}

    for d in data:
        data_grouped[d.products].append(d)

        num_products = len(d.products)
        if num_products > 1:
            num_products_count[num_products] += 1

            if num_products not in num_products_example:
                assert d.mapped_reaction_smiles is not None
                num_products_example[num_products] = d.mapped_reaction_smiles

    for num_products, count in sorted(num_products_count.items()):
        logger.warning(
            f"Found {count} samples with {num_products} products "
            f"(example reaction SMILES: {num_products_example[num_products]})"
        )

    logger.info(f"Got {len(data_grouped)} groups after grouping by product\n")

    fold_to_follow_to_products: dict[DataFold, set[Molecule]] = defaultdict(set)
    if config.dataset_to_follow_dir is not None:
        raw_data_to_follow = load_raw_reactions_files(config.dataset_to_follow_dir)

        for fold in DataFold:
            logger.info(
                f"Read {len(raw_data_to_follow[fold])} samples to follow for fold {fold.value}"
            )

            data_to_follow = parse_samples(
                raw_data_to_follow[fold], num_processes=config.num_processes
            )
            logger.info(f"Left with {len(data_to_follow)} valid samples")

            for d in data_to_follow:
                if len(d.products) > 1:
                    raise ValueError(
                        "Dataset to follow has side-products, which can cause weird results."
                    )

                [product] = list(d.products)
                fold_to_follow_to_products[fold].add(product)

    data_split: dict[DataFold, list[ReactionSample]] = defaultdict(list)
    data_groups_left: list[tuple[Bag[Molecule], list[ReactionSample]]] = []

    for products, d_group in data_grouped.items():
        # Go through folds in reversed order. If one side-product occurrs in the pre-split dataset
        # in test, and another one in train, then the sample goes into test fold of the new dataset.
        for fold in reversed(DataFold):
            if any(product in fold_to_follow_to_products[fold] for product in products):
                data_split[fold].extend(d_group)
                break
        else:
            data_groups_left.append((products, d_group))

    if config.dataset_to_follow_dir is not None:
        fold_sizes_joined = "\n".join(
            [f"{fold.value}: {len(data_split[fold])}" for fold in DataFold]
        )
        logger.info(f"Fold sizes after including the overlapping samples:\n{fold_sizes_joined}")
    else:
        assert not data_split

    if any(len(data_split[fold]) > fold_target_size[fold] for fold in DataFold):
        raise ValueError(
            "After including overlapping sizes some folds are already bigger than desired"
        )

    random.shuffle(data_groups_left)

    folds = list(DataFold)
    current_fold = DataFold.TRAIN

    for _, d_group in data_groups_left:
        if (
            current_fold != DataFold.TEST
            and len(data_split[current_fold]) + len(d_group) > fold_target_size[current_fold]
        ):
            # If the current fold is full progress to the next one.
            current_fold = folds[folds.index(current_fold) + 1]

        data_split[current_fold].extend(d_group)

    fold_sizes_joined = "\n".join(
        [
            f"{fold.value}: {len(data_split[fold])} (target {fold_target_size[fold]})"
            for fold in DataFold
        ]
    )
    logger.info(f"Final fold sizes:\n{fold_sizes_joined}")

    logger.info(f"Saving data under {config.output_dir}")
    for fold, datapoints in data_split.items():
        random.shuffle(datapoints)

        with open(Path(config.output_dir) / f"pista_{fold.value}.smi", "wt") as f:
            for datapoint in datapoints:
                original_str = datapoint.metadata["original_str"]  # type: ignore[typeddict-item]
                f.write(f"{original_str}\n")


def main() -> None:
    set_random_seed(0)
    silence_rdkit_warnings()

    config: SplitDatasetConfig = cli_get_config(argv=None, config_cls=SplitDatasetConfig)
    logger.info(f"Splitting dataset with the following config: {config}")

    split_dataset(config)


if __name__ == "__main__":
    main()
